import os
import json
import socket as socket_module
import sys
import threading
import traceback
import queue

__file__ = os.path.abspath(__file__)
if __file__.endswith((".pyc", ".pyo")):
    __file__ = __file__[:-1]

LOG_FORMAT = "ROBOT: %(asctime)s UTC pid: %(process)d - %(threadName)s - %(levelname)s - %(name)s\n%(message)s\n\n"


def connect(port):
    from robotframework_ls.options import DEFAULT_TIMEOUT
    from robotframework_ls.impl.robot_lsp_constants import ENV_OPTION_ROBOT_DAP_TIMEOUT
    from robocorp_ls_core.robotframework_log import get_logger

    log = get_logger("robotframework_debug_adapter.run_robot__main__.py")

    #  Set TCP keepalive on an open socket.
    #  It activates after 1 second (TCP_KEEPIDLE,) of idleness,
    #  then sends a keepalive ping once every 3 seconds (TCP_KEEPINTVL),
    #  and closes the connection after 5 failed ping (TCP_KEEPCNT), or 15 seconds
    s = socket_module.socket(socket_module.AF_INET, socket_module.SOCK_STREAM)
    try:
        IPPROTO_TCP, SO_KEEPALIVE, TCP_KEEPIDLE, TCP_KEEPINTVL, TCP_KEEPCNT = (
            socket_module.IPPROTO_TCP,
            socket_module.SO_KEEPALIVE,
            socket_module.TCP_KEEPIDLE,  # @UndefinedVariable
            socket_module.TCP_KEEPINTVL,  # @UndefinedVariable
            socket_module.TCP_KEEPCNT,  # @UndefinedVariable
        )
        s.setsockopt(socket_module.SOL_SOCKET, SO_KEEPALIVE, 1)
        s.setsockopt(IPPROTO_TCP, TCP_KEEPIDLE, 1)
        s.setsockopt(IPPROTO_TCP, TCP_KEEPINTVL, 3)
        s.setsockopt(IPPROTO_TCP, TCP_KEEPCNT, 5)
    except AttributeError:
        pass  # May not be available everywhere.

    try:
        # 10 seconds default timeout
        timeout = os.environ.get(ENV_OPTION_ROBOT_DAP_TIMEOUT, DEFAULT_TIMEOUT)
        if timeout is not None:
            s.settimeout(int(timeout))
        s.connect(("127.0.0.1", port))
        s.settimeout(None)  # no timeout after connected
        log.info("Connected.")
        return s
    except:
        log.exception("Could not connect to: %s", port)
        raise


class _RobotTargetComm(threading.Thread):
    def __init__(self, socket, debug: bool) -> None:
        """
        :param socket:
        :param debug:
            True means that we should run in debug mode and False means that the
            --nodebug flag was passed.
        """
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import BaseSchema
        from typing import Union

        threading.Thread.__init__(self)
        self.daemon = True
        self._socket = socket
        self._write_queue: "queue.Queue[Union[BaseSchema, dict, str]]" = queue.Queue()
        self.configuration_done = threading.Event()
        self.terminated = threading.Event()
        self._run_in_debug_mode = debug

        log = get_log()
        log.debug("Patching execution context...")

        from robotframework_debug_adapter.debugger_impl import (
            install_robot_debugger,
        )

        try:
            import robot
        except ImportError:
            log.info("Unable to import Robot (debug will not be available).")
            # If unable to import robot, don't error here (proceed as if
            # it was without debugging -- it should fail later on when
            # about to run the code, at which point the actual DAP is
            # in place).
            self._debugger_impl = None
        else:
            debugger_impl = install_robot_debugger()
            debugger_impl.busy_wait.before_wait.append(self._notify_stopped)
            debugger_impl.write_message = self.write_message  # type: ignore

            log.debug("Finished patching execution context.")
            self._debugger_impl = debugger_impl

        if self._debugger_impl and not self._run_in_debug_mode:
            # If not running in debug mode (in which case we want the stack
            # handling and logging but not any breakpoints/stopping), make
            # sure we disable breaking on logging entries.
            self._debugger_impl.enable_no_debug_mode()

    def _notify_stopped(self):
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import StoppedEvent
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import StoppedEventBody

        thread_id = self.get_current_thread_id()

        exc_name = self._debugger_impl.exc_name
        exc_desc = self._debugger_impl.exc_description

        reason = self._debugger_impl.stop_reason
        body = StoppedEventBody(
            reason.value,
            allThreadsStopped=True,
            threadId=thread_id,
            text=exc_name,
            description=exc_desc,
        )
        msg = StoppedEvent(body)
        self.write_message(msg)

    def start_communication_threads(self, mark_as_pydevd_threads):
        from robocorp_ls_core.debug_adapter_core.debug_adapter_threads import (
            writer_thread,
        )
        from robocorp_ls_core.debug_adapter_core.debug_adapter_threads import (
            reader_thread,
        )

        read_from = self._socket.makefile("rb")
        write_to = self._socket.makefile("wb")

        writer = self._writer_thread = threading.Thread(
            target=writer_thread,
            args=(write_to, self._write_queue, "write to dap", True),
            name="Write from robot to dap (_RobotTargetComm)",
        )
        writer.daemon = True

        reader = self._reader_thread = threading.Thread(
            target=reader_thread,
            args=(
                read_from,
                self.process_message,
                self._write_queue,
                b"read from dap",
                True,
            ),
            name="Read from dap to robot (_RobotTargetComm)",
        )
        reader.daemon = True

        if mark_as_pydevd_threads:
            import pydevd

            pydevd.mark_as_pydevd_daemon_thread(reader)
            pydevd.mark_as_pydevd_daemon_thread(writer)

        reader.start()
        writer.start()
        return reader, writer

    def terminate(self):
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import TerminatedEvent
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            TerminatedEventBody,
        )

        self.write_message(TerminatedEvent(TerminatedEventBody()))

    def write_message(self, msg):
        self._write_queue.put(msg)

    def process_message(self, protocol_message):
        from robotframework_debug_adapter.constants import DEBUG
        from robocorp_ls_core.debug_adapter_core.debug_adapter_threads import (
            READER_THREAD_STOPPED,
        )

        log = get_log()
        if protocol_message is READER_THREAD_STOPPED:
            if DEBUG:
                log.debug("%s: READER_THREAD_STOPPED." % (self.__class__.__name__,))
            return

        if DEBUG:
            log.debug(
                "Process json: %s\n"
                % (json.dumps(protocol_message.to_dict(), indent=4, sort_keys=True),)
            )

        if protocol_message.type == "request":
            method_name = "on_%s_request" % (protocol_message.command,)

        elif protocol_message.type == "event":
            method_name = "on_%s_event" % (protocol_message.event,)

        else:
            if DEBUG:
                log.debug(
                    "Unable to decide how to deal with protocol type: %s in %s.\n"
                    % (protocol_message.type, self.__class__.__name__)
                )
            return

        on_request = getattr(self, method_name, None)
        if on_request is not None:
            on_request(protocol_message)
        else:
            if DEBUG:
                log.debug(
                    "Unhandled: %s not available in %s.\n"
                    % (method_name, self.__class__.__name__)
                )

        # Note: if there's some exception, let it be processed in the caller
        # as the reader_thread does handle it properly.

    def on_terminated_event(self, event):
        self.terminated.set()

    def on_initialize_request(self, request):
        """
        :param InitializeRequest request:
        """
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import InitializedEvent
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import ProcessEvent
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import ProcessEventBody

        # : :type initialize_response: InitializeResponse
        # : :type capabilities: Capabilities
        self._initialize_request_arguments = request.arguments
        initialize_response = build_response(request)
        capabilities = initialize_response.body
        capabilities.supportsConfigurationDoneRequest = True
        capabilities.supportsConditionalBreakpoints = True
        capabilities.supportsEvaluateForHovers = True
        capabilities.supportsHitConditionalBreakpoints = True
        capabilities.supportsLogPoints = True
        capabilities.exceptionBreakpointFilters = [
            {"filter": "logFailure", "label": "Robot Log FAIL", "default": True},
            {"filter": "logError", "label": "Robot Log ERROR", "default": True},
        ]
        # capabilities.supportsSetVariable = True
        self.write_message(initialize_response)
        self.write_message(
            ProcessEvent(ProcessEventBody(sys.executable, systemProcessId=os.getpid()))
        )
        self.write_message(InitializedEvent())

    def on_attach_request(self, request):
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )

        attach_response = build_response(request)
        self.write_message(attach_response)

    def on_setExceptionBreakpoints_request(self, request):
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            SetExceptionBreakpointsArguments,
        )
        from robocorp_ls_core.debug_adapter_core.dap import dap_base_schema

        arguments: SetExceptionBreakpointsArguments = request.arguments
        filters = arguments.filters
        break_on_log_failure = "logFailure" in filters
        break_on_log_error = "logError" in filters

        if self._debugger_impl and self._run_in_debug_mode:
            self._debugger_impl.break_on_log_failure = break_on_log_failure
            self._debugger_impl.break_on_log_error = break_on_log_error
        else:
            if break_on_log_failure or break_on_log_error:
                get_log().info("Unable to break on failures/errors (no debug mode).")

        # Note: no body needed.
        self.write_message(dap_base_schema.build_response(request))

    def on_setBreakpoints_request(self, request):
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import SourceBreakpoint
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import Breakpoint
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            SetBreakpointsResponseBody,
        )
        from robocorp_ls_core.debug_adapter_core.dap import dap_base_schema
        from robotframework_debug_adapter import file_utils
        from robotframework_debug_adapter.debugger_impl import RobotBreakpoint
        from robocorp_ls_core.robotframework_log import get_logger

        log = get_logger("robotframework_debug_adapter.run_robot__main__.py")

        # Just acknowledge that no breakpoints are valid.

        breakpoints = []
        robot_breakpoints = []
        source = request.arguments.source
        path = source.path
        filename = file_utils.norm_file_to_server(path)
        log.info("Normalized %s to %s", path, filename)

        if request.arguments.breakpoints:

            for bp in request.arguments.breakpoints:
                source_breakpoint = SourceBreakpoint(**bp)
                breakpoints.append(
                    Breakpoint(
                        verified=True, line=source_breakpoint.line, source=source
                    ).to_dict()
                )
                hit_condition = None
                try:
                    if source_breakpoint.hitCondition:
                        hit_condition = int(source_breakpoint.hitCondition)
                except:
                    log.exception(
                        "Unable to evaluate hit condition (%s) to an int. Ignoring it.",
                        source_breakpoint.hitCondition,
                    )
                robot_breakpoints.append(
                    RobotBreakpoint(
                        source_breakpoint.line,
                        source_breakpoint.condition,
                        hit_condition,
                        source_breakpoint.logMessage,
                    )
                )

        if self._debugger_impl and self._run_in_debug_mode:
            self._debugger_impl.set_breakpoints(filename, robot_breakpoints)
        else:
            if robot_breakpoints:
                get_log().info("Unable to set breakpoints (no debug mode).")

        self.write_message(
            dap_base_schema.build_response(
                request,
                kwargs=dict(body=SetBreakpointsResponseBody(breakpoints=breakpoints)),
            )
        )

    def on_continue_request(self, request):
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            ContinueResponseBody,
        )

        response = build_response(
            request, kwargs=dict(body=ContinueResponseBody(allThreadsContinued=True))
        )

        if self._debugger_impl and self._run_in_debug_mode:
            self._debugger_impl.step_continue()
        else:
            get_log().info("Unable to continue (no debug mode).")

        self.write_message(response)

    def on_stepIn_request(self, request):
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )

        response = build_response(request)

        if self._debugger_impl and self._run_in_debug_mode:
            self._debugger_impl.step_in()
        else:
            get_log().info("Unable to step in (no debug mode).")
        self.write_message(response)

    def on_next_request(self, request):
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )

        response = build_response(request)

        if self._debugger_impl and self._run_in_debug_mode:
            self._debugger_impl.step_next()
        else:
            get_log().info("Unable to step next (no debug mode).")

        self.write_message(response)

    def on_stepOut_request(self, request):
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )

        response = build_response(request)

        if self._debugger_impl and self._run_in_debug_mode:
            self._debugger_impl.step_out()
        else:
            get_log().info("Unable to step out (no debug mode).")

        self.write_message(response)

    def on_threads_request(self, request):
        """
        :param ThreadsRequest request:
        """
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import Thread
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            ThreadsResponseBody,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )

        thread_id = self.get_current_thread_id()

        threads = [Thread(thread_id, "Main Thread").to_dict()]
        kwargs = {"body": ThreadsResponseBody(threads)}
        # : :type threads_response: ThreadsResponse
        threads_response = build_response(request, kwargs)
        self.write_message(threads_response)

    def get_current_thread_id(self):
        if self._run_in_debug_mode:
            from robotframework_debug_adapter.vendored import force_pydevd  # noqa
            from _pydevd_bundle.pydevd_constants import get_current_thread_id

            return get_current_thread_id(threading.current_thread())
        else:
            return 1

    def on_stackTrace_request(self, request):
        """
        :param StackTraceRequest request:
        """
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            StackTraceResponseBody,
        )

        thread_id = request.arguments.threadId

        if self._debugger_impl and self._run_in_debug_mode:
            frames = self._debugger_impl.get_frames(thread_id)
        else:
            frames = []
            get_log().info("Unable to get stack trace (no debug mode).")

        body = StackTraceResponseBody(stackFrames=frames if frames else [])
        response = build_response(request, kwargs=dict(body=body))
        self.write_message(response)

    def on_configurationDone_request(self, request):
        """
        :param ConfigurationDoneRequest request:
        """
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )

        response = build_response(request)
        self.write_message(response)
        self.configuration_done.set()

    def on_scopes_request(self, request):
        """
        :param ScopesRequest request:
        """
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            ScopesResponseBody,
        )

        frame_id = request.arguments.frameId

        if self._debugger_impl and self._run_in_debug_mode:
            scopes = self._debugger_impl.get_scopes(frame_id)
        else:
            scopes = []
            get_log().info("Unable to step in (no debug mode).")

        body = ScopesResponseBody(scopes if scopes else [])
        response = build_response(request, kwargs=dict(body=body))
        self.write_message(response)

    def on_variables_request(self, request):
        """
        :param VariablesRequest request:
        """
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            VariablesResponseBody,
        )

        variables_reference = request.arguments.variablesReference

        if self._debugger_impl and self._run_in_debug_mode:
            variables = self._debugger_impl.get_variables(variables_reference)
        else:
            variables = []
            get_log().info("Unable to step in (no debug mode).")

        body = VariablesResponseBody(variables if variables else [])
        response = build_response(request, kwargs=dict(body=body))
        self.write_message(response)

    def _evaluate_response(self, request, result, error_message=""):
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import (
            EvaluateResponseBody,
        )
        from robocorp_ls_core.debug_adapter_core.dap.dap_base_schema import (
            build_response,
        )

        body = EvaluateResponseBody(result=result, variablesReference=0)
        if not error_message:
            return build_response(request, kwargs={"body": body})
        else:
            response = build_response(
                request,
                kwargs={"body": body, "success": False, "message": error_message},
            )
            return response

    def on_evaluate_request(self, request):
        from robocorp_ls_core.debug_adapter_core.dap.dap_schema import EvaluateArguments

        arguments: EvaluateArguments = request.arguments

        frame_id = arguments.frameId
        expression = arguments.expression
        context = arguments.context
        if self._debugger_impl and self._run_in_debug_mode:
            eval_info = self._debugger_impl.evaluate(frame_id, expression, context)
            try:
                result = eval_info.future.result()
            except Exception as e:
                err = "".join(traceback.format_exception_only(type(e), e))
                response = self._evaluate_response(request, err, error_message=err)
            else:
                response = self._evaluate_response(request, str(result))
        else:
            get_log().info("Unable to evaluate (no debug mode).")
            response = self._evaluate_response(request, "")

        self.write_message(response)


def get_log():
    from robocorp_ls_core.robotframework_log import get_logger

    return get_logger("robotframework_debug_adapter.run_robot__main__.py")


def main():
    src_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    try:
        import robotframework_ls
    except ImportError:
        # Automatically add it to the path if __main__ is being executed.
        assert os.path.exists(src_folder), "Expected: %s to exist" % (src_folder,)
        sys.path.append(src_folder)
        import robotframework_ls  # @UnusedImport
    robotframework_ls.import_robocorp_ls_core()

    from robocorp_ls_core.robotframework_log import (
        configure_logger,
        log_args_and_python,
        close_logging_streams,
    )

    from robotframework_debug_adapter.constants import LOG_FILENAME
    from robotframework_debug_adapter.constants import LOG_LEVEL

    configure_logger("robot", LOG_LEVEL, LOG_FILENAME)
    log = get_log()
    log_args_and_python(log, sys.argv, robotframework_ls)

    from robotframework_ls.options import DEFAULT_TIMEOUT

    args = sys.argv[1:]
    assert args[0] == "--port"
    port = int(args[1])
    debug = True if args[2] == "--debug" else False

    robot_args = args[3:]

    # Always add the listener (because even when not debugging we want
    # to be able to provide output messages for logging as well as
    # information to be show in the `Robot Output` View).
    robot_args = [
        "--listener=robotframework_debug_adapter.listeners.DebugListener",
        "--listener=robotframework_debug_adapter.listeners.DebugListenerV2",
    ] + robot_args

    s = connect(port)

    if debug:
        if LOG_FILENAME:
            if not os.getenv("PYDEVD_DEBUG_FILE"):
                path, ext = os.path.splitext(LOG_FILENAME)
                os.environ["PYDEVD_DEBUG_FILE"] = f"{path}.pydevd{ext}"
            os.environ["PYDEVD_DEBUG"] = "1"

        # Make sure that we can use pydevd (initialize only in debug mode).
        import robotframework_debug_adapter.vendored.force_pydevd  # @UnusedImport

        import pydevd

        pydevd.settrace(
            "127.0.0.1",
            port=port,
            suspend=False,
            trace_only_current_thread=False,
            overwrite_prev_trace=True,
            patch_multiprocessing=False,
            block_until_connected=True,
            wait_for_ready_to_run=False,
            notify_stdin=False,
        )

    processor = _RobotTargetComm(s, debug=debug)

    from robotframework_debug_adapter import global_vars

    global_vars.set_global_robot_target_comm(processor)

    reader, writer = processor.start_communication_threads(debug)

    if not processor.configuration_done.wait(DEFAULT_TIMEOUT):
        sys.stderr.write(
            "Process not configured for launch in the available timeout.\n"
        )
        sys.exit(1)

    try:
        try:
            import robot
        except ImportError:
            sys.stderr.write("\nError importing robot.\n")
            sys.stderr.write("Python executable: %s.\n\n" % (sys.executable,))
            raise

        from robotframework_debug_adapter.listeners import install_rf_stream_connection

        install_rf_stream_connection(processor.write_message)

        from robot import run_cli

        exitcode = run_cli(robot_args, exit=False)
    finally:
        processor.terminate()
        if processor.terminated.wait(2):
            log.debug("Processed dap terminate event in robot.")
        close_logging_streams()

    sys.exit(exitcode)


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Dict, Iterator, List, Sequence
import json
import itertools
import string
import datetime
from typing import Optional, Callable
import os
import traceback
from contextlib import contextmanager
import sys


_valid_chars = tuple(string.ascii_letters + string.digits)


def _gen_id(level: int = 1) -> Iterator[str]:
    iter_in = tuple(_valid_chars for _ in range(level))
    for entry in itertools.product(*iter_in):
        yield "".join(entry)

    # Recursively generate ids...
    yield from _gen_id(level + 1)


class _Config:
    # Loaded from constructor args
    output_dir: Optional[str]
    max_file_size_in_bytes: int
    max_files: int
    log_html: Optional[str]
    uuid: str

    # Loaded from constructor kwargs (to be used
    # only when used as an API).
    write: Optional[Callable[[str], None]] = None
    initial_time: Optional[datetime.datetime] = None
    additional_info: Optional[Sequence[str]] = None

    def __init__(self, uuid=None):
        if uuid is None:
            import uuid as uuid_module

            uuid = f"{uuid_module.uuid1()}"
        self.uuid = uuid
        self.output_dir = "./out_"


class _RotateHandler:
    def __init__(self, max_file_size_in_bytes: int, max_files: int):
        if max_files <= 0:
            raise ValueError(f"max_files must be > 0. Found: {max_files}")

        self._max_file_size_in_bytes = max_file_size_in_bytes
        self._total_bytes = 0

        self._found_files: List[Path] = []
        self._max_files = max_files

    def rotate_after(self, in_bytes: bytes):
        self._total_bytes += len(in_bytes)
        if self._total_bytes >= self._max_file_size_in_bytes:
            self._total_bytes = 0
            return True
        return False

    def register_file(self, filepath: Path):
        self._found_files.append(filepath)

        while len(self._found_files) > self._max_files:
            p: Path = self._found_files.pop(0)
            try:
                os.remove(p)
            except:
                traceback.print_exc()

    def iter_found_files(self):
        yield from iter(self._found_files)


class _StackEntry:
    def __init__(self, entry_type, entry_id, msg_type, new_msg_type):
        self.entry_type = entry_type
        self.entry_id = entry_id
        self.msg_type = msg_type
        self.new_msg_type = new_msg_type
        self._messages: List[str] = []

    def append(self, msg: str):
        self._messages.append(msg)

    def translated_messages(self) -> Iterator[str]:
        iter_in = iter(self._messages)
        for msg in iter_in:
            if msg.startswith(self.msg_type):
                msg = msg.replace(self.msg_type, self.new_msg_type, 1)
            yield msg


class _StackHandler:
    def __init__(self):
        self._queue = []
        self.recording_writes = False
        self._record_to = None

    @contextmanager
    def push_record(self, entry_type, entry_id, msg_type, new_msg_type):
        self.recording_writes = True
        self._record_to = _StackEntry(entry_type, entry_id, msg_type, new_msg_type)
        try:
            yield
        finally:
            self._queue.append(self._record_to)
            self.recording_writes = False

    def record_msg(self, msg):
        self._record_to.append(msg)

    def pop(self, entry_type, entry_id):
        if not self._queue:
            sys.stderr.write(
                f"RFStream Warning: unable to pop {entry_type} - {entry_id} (empty queue).\n"
            )
        else:
            stack_entry = self._queue[-1]
            if (
                stack_entry.entry_type == entry_type
                and stack_entry.entry_id == entry_id
            ):
                self._queue.pop(-1)
            else:
                for i, stack_entry in enumerate(reversed(self._queue)):
                    if (
                        stack_entry.entry_type == entry_type
                        and stack_entry.entry_id == entry_id
                    ):
                        for _ in range(i):
                            stack_entry = self._queue.pop(-1)
                            sys.stderr.write(
                                f"RFStream Warning: {stack_entry.entry_type} - {stack_entry.entry_id} did not have a corresponding pop.\n"
                            )

                        # The current one (which is a match).
                        self._queue.pop(-1)
                        return

                if entry_type in ("test", "suite"):
                    for i, stack_entry in enumerate(reversed(self._queue)):
                        if stack_entry.entry_type == entry_type:
                            for _ in range(i):
                                stack_entry = self._queue.pop(-1)
                                sys.stderr.write(
                                    f"RFStream Warning: {stack_entry.entry_type} - {stack_entry.entry_id} did not have a corresponding pop.\n"
                                )

                            # The current one (which is a partial match).
                            stack_entry = self._queue.pop(-1)
                            sys.stderr.write(
                                f"RFStream Warning: {stack_entry.entry_type} - {stack_entry.entry_id} pop just by type. Actual request: {entry_type} - {entry_id}\n"
                            )
                            return

                sys.stderr.write(
                    f"RFStream Warning: unable to pop {entry_type} - {entry_id} because it does not match the current top: {stack_entry.entry_type} - {stack_entry.entry_id}\n"
                )

    def __iter__(self):
        for stack_entry in self._queue:
            assert isinstance(stack_entry, _StackEntry)
            yield stack_entry


class _RobotOutputImpl:
    def __init__(self, config: _Config):
        self._written_initial = False

        # Base memory for all streams (rotated or not)
        self._base_memo: Dict[str, str] = {}

        # Memory just for the current stream (if a name is not
        # here it has to be added because the output was rotated).
        self._current_memo: Dict[str, str] = {}

        self._config = config
        self._id = config.uuid

        if config.output_dir is None:
            self._output_dir = None
        else:
            self._output_dir = Path(config.output_dir)
            self._output_dir.mkdir(exist_ok=True)
        self._write = config.write

        self._move_old_runs()

        self._current_entry = 0
        self._current_file: Optional[Path] = None
        self._stream = None

        if config.initial_time is None:
            self._initial_time = datetime.datetime.now()
        else:
            self._initial_time = config.initial_time

        self._stack_handler = _StackHandler()

        self._rotate_handler = _RotateHandler(
            config.max_file_size_in_bytes, config.max_files
        )
        self._id_generator = _gen_id()

        if self._output_dir is not None:
            self._rotate_output()
        else:
            self._current_entry = 1
            self._write_on_start_or_after_rotate()

    @property
    def current_file(self) -> Optional[Path]:
        return self._current_file

    @property
    def initial_time(self) -> datetime.datetime:
        return self._initial_time

    def _rotate_output(self):
        if self._output_dir is not None:
            self._current_memo = {}

            self._current_entry += 1
            if self._current_entry != 1:
                self._current_file = (
                    self._output_dir / f"output_{self._current_entry}.rfstream"
                )
            else:
                self._current_file = self._output_dir / f"output.rfstream"

            if self._stream is not None:
                self._stream.close()
                self._stream = None

            self._rotate_handler.register_file(self._current_file)

            self._stream = self._current_file.open("wb")
            self._write_on_start_or_after_rotate()

    def _write_on_start_or_after_rotate(self):
        if self._current_file is not None:
            print("RFStream output: ", self._current_file.absolute())

        self._write_json("V ", 1)
        self._write_with_separator(
            "T ", (self._initial_time.isoformat(timespec="milliseconds"),)
        )
        self._write_with_separator("ID ", (f"{self._current_entry}", self._id))
        if self._config.additional_info:
            for info in self._config.additional_info:
                self._write_json("I ", info)
        else:
            self._write_json("I ", f"sys.platform={sys.platform}")
            self._write_json("I ", f"python={sys.version}")
            import robot

            robot_version = robot.get_version()
            self._write_json("I ", f"robot={robot_version}")

        for stack_entry in self._stack_handler:
            for msg in stack_entry.translated_messages():
                self._do_write(msg)

    def _do_write(self, s: str) -> None:
        if self._stack_handler.recording_writes:
            self._stack_handler.record_msg(s)

        if self._write is not None:
            self._write(s)

        in_bytes = s.encode("utf-8", errors="replace")
        if self._stream is not None:
            self._stream.write(in_bytes)
            self._stream.flush()
        if self._rotate_handler.rotate_after(in_bytes):
            self._rotate_output()

    def _write_json(self, msg_type, args):
        args_as_str = json.dumps(args)
        s = f"{msg_type}{args_as_str}\n"
        self._do_write(s)
        return s

    def _write_with_separator(self, msg_type, args):
        args_as_str = "|".join(args)
        s = f"{msg_type}{args_as_str}\n"
        self._do_write(s)
        return s

    def get_time_delta(self) -> float:
        delta = datetime.datetime.now() - self._initial_time
        return round(delta.total_seconds(), 3)

    def _move_old_runs(self):
        pass
        # TODO: Handle old runs (move to old runs).
        # for entry in self._output_dir.iterdir():
        #     print(entry)

    def _gen_id(self) -> str:
        while True:
            gen = next(self._id_generator)
            if gen not in self._base_memo:
                return gen

    def _obtain_id(self, s: str) -> str:
        curr_id = self._current_memo.get(s)
        if curr_id is not None:
            if self._stack_handler.recording_writes:
                self._stack_handler.record_msg(f"M {curr_id}:{json.dumps(s)}\n")
            return curr_id

        curr_id = self._base_memo.get(s)
        if curr_id is not None:
            self._write_json(f"M {curr_id}:", s)
            self._current_memo[s] = curr_id
            return curr_id

        new_id = self._gen_id()
        self._write_json(f"M {new_id}:", s)
        self._base_memo[s] = new_id
        self._current_memo[s] = new_id
        return new_id

    def _number(self, v):
        return str(v)

    def start_suite(self, name, suite_id, suite_source, time_delta):
        oid = self._obtain_id
        with self._stack_handler.push_record("suite", suite_id, "SS", "RS"):
            self._write_with_separator(
                "SS ",
                [
                    oid(name),
                    oid(suite_id),
                    oid(suite_source),
                    self._number(time_delta),
                ],
            )

    def end_suite(self, suite_id, status, time_delta):
        oid = self._obtain_id
        self._write_with_separator(
            "ES ",
            [
                oid(status),
                self._number(time_delta),
            ],
        )
        self._stack_handler.pop("suite", suite_id)

    def start_test(self, name, test_id, test_line, time_delta, tags):
        oid = self._obtain_id
        with self._stack_handler.push_record("test", test_id, "ST", "RT"):
            self._write_with_separator(
                "ST ",
                [
                    oid(name),
                    oid(test_id),
                    self._number(test_line),
                    self._number(time_delta),
                ],
            )

            if tags:
                for tag in tags:
                    self.send_tag(tag)

    def send_tag(self, tag: str):
        oid = self._obtain_id
        self._write_with_separator(
            "TG ",
            [
                oid(tag),
            ],
        )

    def send_info(self, info: str):
        self._write_json("I ", info)

    def send_start_time_delta(self, time_delta_in_seconds: float):
        self._write_with_separator("S ", (self._number(time_delta_in_seconds),))

    def end_test(self, test_id, status, message, time_delta):
        oid = self._obtain_id
        self._write_with_separator(
            "ET ",
            [
                oid(status),
                oid(message),
                self._number(time_delta),
            ],
        )
        self._stack_handler.pop("test", test_id)

    def start_keyword(
        self,
        name,
        libname,
        keyword_type,
        doc,
        source,
        lineno,
        start_time_delta,
        args,
        assigns,
    ):
        keyword_type = keyword_type.upper()
        oid = self._obtain_id
        keyword_id = f"{libname}.{name}"
        with self._stack_handler.push_record("keyword", keyword_id, "SK", "RK"):
            self._write_with_separator(
                "SK ",
                [
                    oid(name),
                    oid(libname),
                    oid(keyword_type),
                    oid(doc),
                    oid(source),
                    self._number(lineno),
                    self._number(start_time_delta),
                ],
            )

            if assigns:
                for assign in assigns:
                    self._write_with_separator(
                        "AS ",
                        [
                            oid(assign),
                        ],
                    )
            if args:
                for arg in args:
                    self._write_with_separator(
                        "KA ",
                        [
                            oid(arg),
                        ],
                    )

    def end_keyword(self, name, libname, status, time_delta):
        keyword_id = f"{libname}.{name}"
        oid = self._obtain_id
        self._write_with_separator(
            "EK ",
            [
                oid(status),
                self._number(time_delta),
            ],
        )
        self._stack_handler.pop("keyword", keyword_id)

    def log_message(self, level, message, time_delta, html):
        oid = self._obtain_id

        msg_type = "L "
        if html in ("true", "yes", 1, True):
            # From output.xml it's "true", from listener it's "yes".
            msg_type = "LH "

        self._write_with_separator(
            msg_type,
            [
                # ERROR = E
                # FAIL = F
                # INFO = I
                # WARN = W
                level[0].upper(),
                oid(message),
                self._number(time_delta),
            ],
        )

    def close(self):
        import base64

        log_html = self._config.log_html
        if log_html:
            target = os.path.abspath(log_html)
            dirname = os.path.dirname(target)
            if not os.path.exists(dirname):
                os.makedirs(dirname, exist_ok=True)

            print(f"RFstream log (html): {target}")

            with open(target, "wb") as stream:
                from robot_out_stream import index

                index_html_contents = index.INDEX_HTML_CONTENTS

                # We want to replace the string so that we write
                sample_contents_index = index_html_contents.index("getSampleContents")
                return_json_parse_index = index_html_contents.index(
                    "return JSON.parse", sample_contents_index
                )

                end_of_json_parse_index = (
                    index_html_contents.index(")}}", return_json_parse_index) + 1
                )

                stream.write(
                    index_html_contents[:return_json_parse_index].encode("utf-8")
                )

                import zlib

                stream.write(b"\nlet chunks = [")
                for f in self._rotate_handler.iter_found_files():
                    with open(f, "rb") as fsrc:
                        while True:
                            chunk = fsrc.read(1024 * 8)
                            if not chunk:
                                break
                            stream.write(b'"')
                            chunk = zlib.compress(chunk)
                            stream.write(base64.b64encode(chunk))
                            stream.write(b'",\n')
                stream.write(b"];\n")

                # Add code to decompress the data we added.
                stream.write(
                    b"""
/*! pako 2.1.0 https://github.com/nodeca/pako @license (MIT AND Zlib) */
!function(e,t){"object"==typeof exports&&"undefined"!=typeof module?t(exports):"function"==typeof define&&define.amd?define(["exports"],t):t((e="undefined"!=typeof globalThis?globalThis:e||self).pako={})}(this,(function(e){"use strict";var t=(e,t,i,n)=>{let a=65535&e|0,r=e>>>16&65535|0,o=0;for(;0!==i;){o=i>2e3?2e3:i,i-=o;do{a=a+t[n++]|0,r=r+a|0}while(--o);a%=65521,r%=65521}return a|r<<16|0};const i=new Uint32Array((()=>{let e,t=[];for(var i=0;i<256;i++){e=i;for(var n=0;n<8;n++)e=1&e?3988292384^e>>>1:e>>>1;t[i]=e}return t})());var n=(e,t,n,a)=>{const r=i,o=a+n;e^=-1;for(let i=a;i<o;i++)e=e>>>8^r[255&(e^t[i])];return-1^e};const a=16209;var r=function(e,t){let i,n,r,o,s,l,d,f,c,h,u,w,b,m,k,_,g,p,v,x,y,E,R,A;const Z=e.state;i=e.next_in,R=e.input,n=i+(e.avail_in-5),r=e.next_out,A=e.output,o=r-(t-e.avail_out),s=r+(e.avail_out-257),l=Z.dmax,d=Z.wsize,f=Z.whave,c=Z.wnext,h=Z.window,u=Z.hold,w=Z.bits,b=Z.lencode,m=Z.distcode,k=(1<<Z.lenbits)-1,_=(1<<Z.distbits)-1;e:do{w<15&&(u+=R[i++]<<w,w+=8,u+=R[i++]<<w,w+=8),g=b[u&k];t:for(;;){if(p=g>>>24,u>>>=p,w-=p,p=g>>>16&255,0===p)A[r++]=65535&g;else{if(!(16&p)){if(0==(64&p)){g=b[(65535&g)+(u&(1<<p)-1)];continue t}if(32&p){Z.mode=16191;break e}e.msg="invalid literal/length code",Z.mode=a;break e}v=65535&g,p&=15,p&&(w<p&&(u+=R[i++]<<w,w+=8),v+=u&(1<<p)-1,u>>>=p,w-=p),w<15&&(u+=R[i++]<<w,w+=8,u+=R[i++]<<w,w+=8),g=m[u&_];i:for(;;){if(p=g>>>24,u>>>=p,w-=p,p=g>>>16&255,!(16&p)){if(0==(64&p)){g=m[(65535&g)+(u&(1<<p)-1)];continue i}e.msg="invalid distance code",Z.mode=a;break e}if(x=65535&g,p&=15,w<p&&(u+=R[i++]<<w,w+=8,w<p&&(u+=R[i++]<<w,w+=8)),x+=u&(1<<p)-1,x>l){e.msg="invalid distance too far back",Z.mode=a;break e}if(u>>>=p,w-=p,p=r-o,x>p){if(p=x-p,p>f&&Z.sane){e.msg="invalid distance too far back",Z.mode=a;break e}if(y=0,E=h,0===c){if(y+=d-p,p<v){v-=p;do{A[r++]=h[y++]}while(--p);y=r-x,E=A}}else if(c<p){if(y+=d+c-p,p-=c,p<v){v-=p;do{A[r++]=h[y++]}while(--p);if(y=0,c<v){p=c,v-=p;do{A[r++]=h[y++]}while(--p);y=r-x,E=A}}}else if(y+=c-p,p<v){v-=p;do{A[r++]=h[y++]}while(--p);y=r-x,E=A}for(;v>2;)A[r++]=E[y++],A[r++]=E[y++],A[r++]=E[y++],v-=3;v&&(A[r++]=E[y++],v>1&&(A[r++]=E[y++]))}else{y=r-x;do{A[r++]=A[y++],A[r++]=A[y++],A[r++]=A[y++],v-=3}while(v>2);v&&(A[r++]=A[y++],v>1&&(A[r++]=A[y++]))}break}}break}}while(i<n&&r<s);v=w>>3,i-=v,w-=v<<3,u&=(1<<w)-1,e.next_in=i,e.next_out=r,e.avail_in=i<n?n-i+5:5-(i-n),e.avail_out=r<s?s-r+257:257-(r-s),Z.hold=u,Z.bits=w};const o=15,s=new Uint16Array([3,4,5,6,7,8,9,10,11,13,15,17,19,23,27,31,35,43,51,59,67,83,99,115,131,163,195,227,258,0,0]),l=new Uint8Array([16,16,16,16,16,16,16,16,17,17,17,17,18,18,18,18,19,19,19,19,20,20,20,20,21,21,21,21,16,72,78]),d=new Uint16Array([1,2,3,4,5,7,9,13,17,25,33,49,65,97,129,193,257,385,513,769,1025,1537,2049,3073,4097,6145,8193,12289,16385,24577,0,0]),f=new Uint8Array([16,16,16,16,17,17,18,18,19,19,20,20,21,21,22,22,23,23,24,24,25,25,26,26,27,27,28,28,29,29,64,64]);var c=(e,t,i,n,a,r,c,h)=>{const u=h.bits;let w,b,m,k,_,g,p=0,v=0,x=0,y=0,E=0,R=0,A=0,Z=0,S=0,T=0,O=null;const U=new Uint16Array(16),D=new Uint16Array(16);let I,B,N,C=null;for(p=0;p<=o;p++)U[p]=0;for(v=0;v<n;v++)U[t[i+v]]++;for(E=u,y=o;y>=1&&0===U[y];y--);if(E>y&&(E=y),0===y)return a[r++]=20971520,a[r++]=20971520,h.bits=1,0;for(x=1;x<y&&0===U[x];x++);for(E<x&&(E=x),Z=1,p=1;p<=o;p++)if(Z<<=1,Z-=U[p],Z<0)return-1;if(Z>0&&(0===e||1!==y))return-1;for(D[1]=0,p=1;p<o;p++)D[p+1]=D[p]+U[p];for(v=0;v<n;v++)0!==t[i+v]&&(c[D[t[i+v]]++]=v);if(0===e?(O=C=c,g=20):1===e?(O=s,C=l,g=257):(O=d,C=f,g=0),T=0,v=0,p=x,_=r,R=E,A=0,m=-1,S=1<<E,k=S-1,1===e&&S>852||2===e&&S>592)return 1;for(;;){I=p-A,c[v]+1<g?(B=0,N=c[v]):c[v]>=g?(B=C[c[v]-g],N=O[c[v]-g]):(B=96,N=0),w=1<<p-A,b=1<<R,x=b;do{b-=w,a[_+(T>>A)+b]=I<<24|B<<16|N|0}while(0!==b);for(w=1<<p-1;T&w;)w>>=1;if(0!==w?(T&=w-1,T+=w):T=0,v++,0==--U[p]){if(p===y)break;p=t[i+c[v]]}if(p>E&&(T&k)!==m){for(0===A&&(A=E),_+=x,R=p-A,Z=1<<R;R+A<y&&(Z-=U[R+A],!(Z<=0));)R++,Z<<=1;if(S+=1<<R,1===e&&S>852||2===e&&S>592)return 1;m=T&k,a[m]=E<<24|R<<16|_-r|0}}return 0!==T&&(a[_+T]=p-A<<24|64<<16|0),h.bits=E,0},h={Z_NO_FLUSH:0,Z_PARTIAL_FLUSH:1,Z_SYNC_FLUSH:2,Z_FULL_FLUSH:3,Z_FINISH:4,Z_BLOCK:5,Z_TREES:6,Z_OK:0,Z_STREAM_END:1,Z_NEED_DICT:2,Z_ERRNO:-1,Z_STREAM_ERROR:-2,Z_DATA_ERROR:-3,Z_MEM_ERROR:-4,Z_BUF_ERROR:-5,Z_NO_COMPRESSION:0,Z_BEST_SPEED:1,Z_BEST_COMPRESSION:9,Z_DEFAULT_COMPRESSION:-1,Z_FILTERED:1,Z_HUFFMAN_ONLY:2,Z_RLE:3,Z_FIXED:4,Z_DEFAULT_STRATEGY:0,Z_BINARY:0,Z_TEXT:1,Z_UNKNOWN:2,Z_DEFLATED:8};const{Z_FINISH:u,Z_BLOCK:w,Z_TREES:b,Z_OK:m,Z_STREAM_END:k,Z_NEED_DICT:_,Z_STREAM_ERROR:g,Z_DATA_ERROR:p,Z_MEM_ERROR:v,Z_BUF_ERROR:x,Z_DEFLATED:y}=h,E=16180,R=16190,A=16191,Z=16192,S=16194,T=16199,O=16200,U=16206,D=16209,I=e=>(e>>>24&255)+(e>>>8&65280)+((65280&e)<<8)+((255&e)<<24);function B(){this.strm=null,this.mode=0,this.last=!1,this.wrap=0,this.havedict=!1,this.flags=0,this.dmax=0,this.check=0,this.total=0,this.head=null,this.wbits=0,this.wsize=0,this.whave=0,this.wnext=0,this.window=null,this.hold=0,this.bits=0,this.length=0,this.offset=0,this.extra=0,this.lencode=null,this.distcode=null,this.lenbits=0,this.distbits=0,this.ncode=0,this.nlen=0,this.ndist=0,this.have=0,this.next=null,this.lens=new Uint16Array(320),this.work=new Uint16Array(288),this.lendyn=null,this.distdyn=null,this.sane=0,this.back=0,this.was=0}const N=e=>{if(!e)return 1;const t=e.state;return!t||t.strm!==e||t.mode<E||t.mode>16211?1:0},C=e=>{if(N(e))return g;const t=e.state;return e.total_in=e.total_out=t.total=0,e.msg="",t.wrap&&(e.adler=1&t.wrap),t.mode=E,t.last=0,t.havedict=0,t.flags=-1,t.dmax=32768,t.head=null,t.hold=0,t.bits=0,t.lencode=t.lendyn=new Int32Array(852),t.distcode=t.distdyn=new Int32Array(592),t.sane=1,t.back=-1,m},z=e=>{if(N(e))return g;const t=e.state;return t.wsize=0,t.whave=0,t.wnext=0,C(e)},F=(e,t)=>{let i;if(N(e))return g;const n=e.state;return t<0?(i=0,t=-t):(i=5+(t>>4),t<48&&(t&=15)),t&&(t<8||t>15)?g:(null!==n.window&&n.wbits!==t&&(n.window=null),n.wrap=i,n.wbits=t,z(e))},L=(e,t)=>{if(!e)return g;const i=new B;e.state=i,i.strm=e,i.window=null,i.mode=E;const n=F(e,t);return n!==m&&(e.state=null),n};let M,H,j=!0;const K=e=>{if(j){M=new Int32Array(512),H=new Int32Array(32);let t=0;for(;t<144;)e.lens[t++]=8;for(;t<256;)e.lens[t++]=9;for(;t<280;)e.lens[t++]=7;for(;t<288;)e.lens[t++]=8;for(c(1,e.lens,0,288,M,0,e.work,{bits:9}),t=0;t<32;)e.lens[t++]=5;c(2,e.lens,0,32,H,0,e.work,{bits:5}),j=!1}e.lencode=M,e.lenbits=9,e.distcode=H,e.distbits=5},P=(e,t,i,n)=>{let a;const r=e.state;return null===r.window&&(r.wsize=1<<r.wbits,r.wnext=0,r.whave=0,r.window=new Uint8Array(r.wsize)),n>=r.wsize?(r.window.set(t.subarray(i-r.wsize,i),0),r.wnext=0,r.whave=r.wsize):(a=r.wsize-r.wnext,a>n&&(a=n),r.window.set(t.subarray(i-n,i-n+a),r.wnext),(n-=a)?(r.window.set(t.subarray(i-n,i),0),r.wnext=n,r.whave=r.wsize):(r.wnext+=a,r.wnext===r.wsize&&(r.wnext=0),r.whave<r.wsize&&(r.whave+=a))),0};var Y={inflateReset:z,inflateReset2:F,inflateResetKeep:C,inflateInit:e=>L(e,15),inflateInit2:L,inflate:(e,i)=>{let a,o,s,l,d,f,h,B,C,z,F,L,M,H,j,Y,G,X,W,q,J,Q,V=0;const $=new Uint8Array(4);let ee,te;const ie=new Uint8Array([16,17,18,0,8,7,9,6,10,5,11,4,12,3,13,2,14,1,15]);if(N(e)||!e.output||!e.input&&0!==e.avail_in)return g;a=e.state,a.mode===A&&(a.mode=Z),d=e.next_out,s=e.output,h=e.avail_out,l=e.next_in,o=e.input,f=e.avail_in,B=a.hold,C=a.bits,z=f,F=h,Q=m;e:for(;;)switch(a.mode){case E:if(0===a.wrap){a.mode=Z;break}for(;C<16;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}if(2&a.wrap&&35615===B){0===a.wbits&&(a.wbits=15),a.check=0,$[0]=255&B,$[1]=B>>>8&255,a.check=n(a.check,$,2,0),B=0,C=0,a.mode=16181;break}if(a.head&&(a.head.done=!1),!(1&a.wrap)||(((255&B)<<8)+(B>>8))%31){e.msg="incorrect header check",a.mode=D;break}if((15&B)!==y){e.msg="unknown compression method",a.mode=D;break}if(B>>>=4,C-=4,J=8+(15&B),0===a.wbits&&(a.wbits=J),J>15||J>a.wbits){e.msg="invalid window size",a.mode=D;break}a.dmax=1<<a.wbits,a.flags=0,e.adler=a.check=1,a.mode=512&B?16189:A,B=0,C=0;break;case 16181:for(;C<16;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}if(a.flags=B,(255&a.flags)!==y){e.msg="unknown compression method",a.mode=D;break}if(57344&a.flags){e.msg="unknown header flags set",a.mode=D;break}a.head&&(a.head.text=B>>8&1),512&a.flags&&4&a.wrap&&($[0]=255&B,$[1]=B>>>8&255,a.check=n(a.check,$,2,0)),B=0,C=0,a.mode=16182;case 16182:for(;C<32;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}a.head&&(a.head.time=B),512&a.flags&&4&a.wrap&&($[0]=255&B,$[1]=B>>>8&255,$[2]=B>>>16&255,$[3]=B>>>24&255,a.check=n(a.check,$,4,0)),B=0,C=0,a.mode=16183;case 16183:for(;C<16;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}a.head&&(a.head.xflags=255&B,a.head.os=B>>8),512&a.flags&&4&a.wrap&&($[0]=255&B,$[1]=B>>>8&255,a.check=n(a.check,$,2,0)),B=0,C=0,a.mode=16184;case 16184:if(1024&a.flags){for(;C<16;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}a.length=B,a.head&&(a.head.extra_len=B),512&a.flags&&4&a.wrap&&($[0]=255&B,$[1]=B>>>8&255,a.check=n(a.check,$,2,0)),B=0,C=0}else a.head&&(a.head.extra=null);a.mode=16185;case 16185:if(1024&a.flags&&(L=a.length,L>f&&(L=f),L&&(a.head&&(J=a.head.extra_len-a.length,a.head.extra||(a.head.extra=new Uint8Array(a.head.extra_len)),a.head.extra.set(o.subarray(l,l+L),J)),512&a.flags&&4&a.wrap&&(a.check=n(a.check,o,L,l)),f-=L,l+=L,a.length-=L),a.length))break e;a.length=0,a.mode=16186;case 16186:if(2048&a.flags){if(0===f)break e;L=0;do{J=o[l+L++],a.head&&J&&a.length<65536&&(a.head.name+=String.fromCharCode(J))}while(J&&L<f);if(512&a.flags&&4&a.wrap&&(a.check=n(a.check,o,L,l)),f-=L,l+=L,J)break e}else a.head&&(a.head.name=null);a.length=0,a.mode=16187;case 16187:if(4096&a.flags){if(0===f)break e;L=0;do{J=o[l+L++],a.head&&J&&a.length<65536&&(a.head.comment+=String.fromCharCode(J))}while(J&&L<f);if(512&a.flags&&4&a.wrap&&(a.check=n(a.check,o,L,l)),f-=L,l+=L,J)break e}else a.head&&(a.head.comment=null);a.mode=16188;case 16188:if(512&a.flags){for(;C<16;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}if(4&a.wrap&&B!==(65535&a.check)){e.msg="header crc mismatch",a.mode=D;break}B=0,C=0}a.head&&(a.head.hcrc=a.flags>>9&1,a.head.done=!0),e.adler=a.check=0,a.mode=A;break;case 16189:for(;C<32;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}e.adler=a.check=I(B),B=0,C=0,a.mode=R;case R:if(0===a.havedict)return e.next_out=d,e.avail_out=h,e.next_in=l,e.avail_in=f,a.hold=B,a.bits=C,_;e.adler=a.check=1,a.mode=A;case A:if(i===w||i===b)break e;case Z:if(a.last){B>>>=7&C,C-=7&C,a.mode=U;break}for(;C<3;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}switch(a.last=1&B,B>>>=1,C-=1,3&B){case 0:a.mode=16193;break;case 1:if(K(a),a.mode=T,i===b){B>>>=2,C-=2;break e}break;case 2:a.mode=16196;break;case 3:e.msg="invalid block type",a.mode=D}B>>>=2,C-=2;break;case 16193:for(B>>>=7&C,C-=7&C;C<32;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}if((65535&B)!=(B>>>16^65535)){e.msg="invalid stored block lengths",a.mode=D;break}if(a.length=65535&B,B=0,C=0,a.mode=S,i===b)break e;case S:a.mode=16195;case 16195:if(L=a.length,L){if(L>f&&(L=f),L>h&&(L=h),0===L)break e;s.set(o.subarray(l,l+L),d),f-=L,l+=L,h-=L,d+=L,a.length-=L;break}a.mode=A;break;case 16196:for(;C<14;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}if(a.nlen=257+(31&B),B>>>=5,C-=5,a.ndist=1+(31&B),B>>>=5,C-=5,a.ncode=4+(15&B),B>>>=4,C-=4,a.nlen>286||a.ndist>30){e.msg="too many length or distance symbols",a.mode=D;break}a.have=0,a.mode=16197;case 16197:for(;a.have<a.ncode;){for(;C<3;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}a.lens[ie[a.have++]]=7&B,B>>>=3,C-=3}for(;a.have<19;)a.lens[ie[a.have++]]=0;if(a.lencode=a.lendyn,a.lenbits=7,ee={bits:a.lenbits},Q=c(0,a.lens,0,19,a.lencode,0,a.work,ee),a.lenbits=ee.bits,Q){e.msg="invalid code lengths set",a.mode=D;break}a.have=0,a.mode=16198;case 16198:for(;a.have<a.nlen+a.ndist;){for(;V=a.lencode[B&(1<<a.lenbits)-1],j=V>>>24,Y=V>>>16&255,G=65535&V,!(j<=C);){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}if(G<16)B>>>=j,C-=j,a.lens[a.have++]=G;else{if(16===G){for(te=j+2;C<te;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}if(B>>>=j,C-=j,0===a.have){e.msg="invalid bit length repeat",a.mode=D;break}J=a.lens[a.have-1],L=3+(3&B),B>>>=2,C-=2}else if(17===G){for(te=j+3;C<te;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}B>>>=j,C-=j,J=0,L=3+(7&B),B>>>=3,C-=3}else{for(te=j+7;C<te;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}B>>>=j,C-=j,J=0,L=11+(127&B),B>>>=7,C-=7}if(a.have+L>a.nlen+a.ndist){e.msg="invalid bit length repeat",a.mode=D;break}for(;L--;)a.lens[a.have++]=J}}if(a.mode===D)break;if(0===a.lens[256]){e.msg="invalid code -- missing end-of-block",a.mode=D;break}if(a.lenbits=9,ee={bits:a.lenbits},Q=c(1,a.lens,0,a.nlen,a.lencode,0,a.work,ee),a.lenbits=ee.bits,Q){e.msg="invalid literal/lengths set",a.mode=D;break}if(a.distbits=6,a.distcode=a.distdyn,ee={bits:a.distbits},Q=c(2,a.lens,a.nlen,a.ndist,a.distcode,0,a.work,ee),a.distbits=ee.bits,Q){e.msg="invalid distances set",a.mode=D;break}if(a.mode=T,i===b)break e;case T:a.mode=O;case O:if(f>=6&&h>=258){e.next_out=d,e.avail_out=h,e.next_in=l,e.avail_in=f,a.hold=B,a.bits=C,r(e,F),d=e.next_out,s=e.output,h=e.avail_out,l=e.next_in,o=e.input,f=e.avail_in,B=a.hold,C=a.bits,a.mode===A&&(a.back=-1);break}for(a.back=0;V=a.lencode[B&(1<<a.lenbits)-1],j=V>>>24,Y=V>>>16&255,G=65535&V,!(j<=C);){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}if(Y&&0==(240&Y)){for(X=j,W=Y,q=G;V=a.lencode[q+((B&(1<<X+W)-1)>>X)],j=V>>>24,Y=V>>>16&255,G=65535&V,!(X+j<=C);){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}B>>>=X,C-=X,a.back+=X}if(B>>>=j,C-=j,a.back+=j,a.length=G,0===Y){a.mode=16205;break}if(32&Y){a.back=-1,a.mode=A;break}if(64&Y){e.msg="invalid literal/length code",a.mode=D;break}a.extra=15&Y,a.mode=16201;case 16201:if(a.extra){for(te=a.extra;C<te;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}a.length+=B&(1<<a.extra)-1,B>>>=a.extra,C-=a.extra,a.back+=a.extra}a.was=a.length,a.mode=16202;case 16202:for(;V=a.distcode[B&(1<<a.distbits)-1],j=V>>>24,Y=V>>>16&255,G=65535&V,!(j<=C);){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}if(0==(240&Y)){for(X=j,W=Y,q=G;V=a.distcode[q+((B&(1<<X+W)-1)>>X)],j=V>>>24,Y=V>>>16&255,G=65535&V,!(X+j<=C);){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}B>>>=X,C-=X,a.back+=X}if(B>>>=j,C-=j,a.back+=j,64&Y){e.msg="invalid distance code",a.mode=D;break}a.offset=G,a.extra=15&Y,a.mode=16203;case 16203:if(a.extra){for(te=a.extra;C<te;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}a.offset+=B&(1<<a.extra)-1,B>>>=a.extra,C-=a.extra,a.back+=a.extra}if(a.offset>a.dmax){e.msg="invalid distance too far back",a.mode=D;break}a.mode=16204;case 16204:if(0===h)break e;if(L=F-h,a.offset>L){if(L=a.offset-L,L>a.whave&&a.sane){e.msg="invalid distance too far back",a.mode=D;break}L>a.wnext?(L-=a.wnext,M=a.wsize-L):M=a.wnext-L,L>a.length&&(L=a.length),H=a.window}else H=s,M=d-a.offset,L=a.length;L>h&&(L=h),h-=L,a.length-=L;do{s[d++]=H[M++]}while(--L);0===a.length&&(a.mode=O);break;case 16205:if(0===h)break e;s[d++]=a.length,h--,a.mode=O;break;case U:if(a.wrap){for(;C<32;){if(0===f)break e;f--,B|=o[l++]<<C,C+=8}if(F-=h,e.total_out+=F,a.total+=F,4&a.wrap&&F&&(e.adler=a.check=a.flags?n(a.check,s,F,d-F):t(a.check,s,F,d-F)),F=h,4&a.wrap&&(a.flags?B:I(B))!==a.check){e.msg="incorrect data check",a.mode=D;break}B=0,C=0}a.mode=16207;case 16207:if(a.wrap&&a.flags){for(;C<32;){if(0===f)break e;f--,B+=o[l++]<<C,C+=8}if(4&a.wrap&&B!==(4294967295&a.total)){e.msg="incorrect length check",a.mode=D;break}B=0,C=0}a.mode=16208;case 16208:Q=k;break e;case D:Q=p;break e;case 16210:return v;default:return g}return e.next_out=d,e.avail_out=h,e.next_in=l,e.avail_in=f,a.hold=B,a.bits=C,(a.wsize||F!==e.avail_out&&a.mode<D&&(a.mode<U||i!==u))&&P(e,e.output,e.next_out,F-e.avail_out),z-=e.avail_in,F-=e.avail_out,e.total_in+=z,e.total_out+=F,a.total+=F,4&a.wrap&&F&&(e.adler=a.check=a.flags?n(a.check,s,F,e.next_out-F):t(a.check,s,F,e.next_out-F)),e.data_type=a.bits+(a.last?64:0)+(a.mode===A?128:0)+(a.mode===T||a.mode===S?256:0),(0===z&&0===F||i===u)&&Q===m&&(Q=x),Q},inflateEnd:e=>{if(N(e))return g;let t=e.state;return t.window&&(t.window=null),e.state=null,m},inflateGetHeader:(e,t)=>{if(N(e))return g;const i=e.state;return 0==(2&i.wrap)?g:(i.head=t,t.done=!1,m)},inflateSetDictionary:(e,i)=>{const n=i.length;let a,r,o;return N(e)?g:(a=e.state,0!==a.wrap&&a.mode!==R?g:a.mode===R&&(r=1,r=t(r,i,n,0),r!==a.check)?p:(o=P(e,i,n,n),o?(a.mode=16210,v):(a.havedict=1,m)))},inflateInfo:"pako inflate (from Nodeca project)"};const G=(e,t)=>Object.prototype.hasOwnProperty.call(e,t);var X=function(e){const t=Array.prototype.slice.call(arguments,1);for(;t.length;){const i=t.shift();if(i){if("object"!=typeof i)throw new TypeError(i+"must be non-object");for(const t in i)G(i,t)&&(e[t]=i[t])}}return e},W=e=>{let t=0;for(let i=0,n=e.length;i<n;i++)t+=e[i].length;const i=new Uint8Array(t);for(let t=0,n=0,a=e.length;t<a;t++){let a=e[t];i.set(a,n),n+=a.length}return i};let q=!0;try{String.fromCharCode.apply(null,new Uint8Array(1))}catch(e){q=!1}const J=new Uint8Array(256);for(let e=0;e<256;e++)J[e]=e>=252?6:e>=248?5:e>=240?4:e>=224?3:e>=192?2:1;J[254]=J[254]=1;var Q=e=>{if("function"==typeof TextEncoder&&TextEncoder.prototype.encode)return(new TextEncoder).encode(e);let t,i,n,a,r,o=e.length,s=0;for(a=0;a<o;a++)i=e.charCodeAt(a),55296==(64512&i)&&a+1<o&&(n=e.charCodeAt(a+1),56320==(64512&n)&&(i=65536+(i-55296<<10)+(n-56320),a++)),s+=i<128?1:i<2048?2:i<65536?3:4;for(t=new Uint8Array(s),r=0,a=0;r<s;a++)i=e.charCodeAt(a),55296==(64512&i)&&a+1<o&&(n=e.charCodeAt(a+1),56320==(64512&n)&&(i=65536+(i-55296<<10)+(n-56320),a++)),i<128?t[r++]=i:i<2048?(t[r++]=192|i>>>6,t[r++]=128|63&i):i<65536?(t[r++]=224|i>>>12,t[r++]=128|i>>>6&63,t[r++]=128|63&i):(t[r++]=240|i>>>18,t[r++]=128|i>>>12&63,t[r++]=128|i>>>6&63,t[r++]=128|63&i);return t},V=(e,t)=>{const i=t||e.length;if("function"==typeof TextDecoder&&TextDecoder.prototype.decode)return(new TextDecoder).decode(e.subarray(0,t));let n,a;const r=new Array(2*i);for(a=0,n=0;n<i;){let t=e[n++];if(t<128){r[a++]=t;continue}let o=J[t];if(o>4)r[a++]=65533,n+=o-1;else{for(t&=2===o?31:3===o?15:7;o>1&&n<i;)t=t<<6|63&e[n++],o--;o>1?r[a++]=65533:t<65536?r[a++]=t:(t-=65536,r[a++]=55296|t>>10&1023,r[a++]=56320|1023&t)}}return((e,t)=>{if(t<65534&&e.subarray&&q)return String.fromCharCode.apply(null,e.length===t?e:e.subarray(0,t));let i="";for(let n=0;n<t;n++)i+=String.fromCharCode(e[n]);return i})(r,a)},$=(e,t)=>{(t=t||e.length)>e.length&&(t=e.length);let i=t-1;for(;i>=0&&128==(192&e[i]);)i--;return i<0||0===i?t:i+J[e[i]]>t?i:t},ee={2:"need dictionary",1:"stream end",0:"","-1":"file error","-2":"stream error","-3":"data error","-4":"insufficient memory","-5":"buffer error","-6":"incompatible version"};var te=function(){this.input=null,this.next_in=0,this.avail_in=0,this.total_in=0,this.output=null,this.next_out=0,this.avail_out=0,this.total_out=0,this.msg="",this.state=null,this.data_type=2,this.adler=0};var ie=function(){this.text=0,this.time=0,this.xflags=0,this.os=0,this.extra=null,this.extra_len=0,this.name="",this.comment="",this.hcrc=0,this.done=!1};const ne=Object.prototype.toString,{Z_NO_FLUSH:ae,Z_FINISH:re,Z_OK:oe,Z_STREAM_END:se,Z_NEED_DICT:le,Z_STREAM_ERROR:de,Z_DATA_ERROR:fe,Z_MEM_ERROR:ce}=h;function he(e){this.options=X({chunkSize:65536,windowBits:15,to:""},e||{});const t=this.options;t.raw&&t.windowBits>=0&&t.windowBits<16&&(t.windowBits=-t.windowBits,0===t.windowBits&&(t.windowBits=-15)),!(t.windowBits>=0&&t.windowBits<16)||e&&e.windowBits||(t.windowBits+=32),t.windowBits>15&&t.windowBits<48&&0==(15&t.windowBits)&&(t.windowBits|=15),this.err=0,this.msg="",this.ended=!1,this.chunks=[],this.strm=new te,this.strm.avail_out=0;let i=Y.inflateInit2(this.strm,t.windowBits);if(i!==oe)throw new Error(ee[i]);if(this.header=new ie,Y.inflateGetHeader(this.strm,this.header),t.dictionary&&("string"==typeof t.dictionary?t.dictionary=Q(t.dictionary):"[object ArrayBuffer]"===ne.call(t.dictionary)&&(t.dictionary=new Uint8Array(t.dictionary)),t.raw&&(i=Y.inflateSetDictionary(this.strm,t.dictionary),i!==oe)))throw new Error(ee[i])}function ue(e,t){const i=new he(t);if(i.push(e),i.err)throw i.msg||ee[i.err];return i.result}he.prototype.push=function(e,t){const i=this.strm,n=this.options.chunkSize,a=this.options.dictionary;let r,o,s;if(this.ended)return!1;for(o=t===~~t?t:!0===t?re:ae,"[object ArrayBuffer]"===ne.call(e)?i.input=new Uint8Array(e):i.input=e,i.next_in=0,i.avail_in=i.input.length;;){for(0===i.avail_out&&(i.output=new Uint8Array(n),i.next_out=0,i.avail_out=n),r=Y.inflate(i,o),r===le&&a&&(r=Y.inflateSetDictionary(i,a),r===oe?r=Y.inflate(i,o):r===fe&&(r=le));i.avail_in>0&&r===se&&i.state.wrap>0&&0!==e[i.next_in];)Y.inflateReset(i),r=Y.inflate(i,o);switch(r){case de:case fe:case le:case ce:return this.onEnd(r),this.ended=!0,!1}if(s=i.avail_out,i.next_out&&(0===i.avail_out||r===se))if("string"===this.options.to){let e=$(i.output,i.next_out),t=i.next_out-e,a=V(i.output,e);i.next_out=t,i.avail_out=n-t,t&&i.output.set(i.output.subarray(e,e+t),0),this.onData(a)}else this.onData(i.output.length===i.next_out?i.output:i.output.subarray(0,i.next_out));if(r!==oe||0!==s){if(r===se)return r=Y.inflateEnd(this.strm),this.onEnd(r),this.ended=!0,!0;if(0===i.avail_in)break}}return!0},he.prototype.onData=function(e){this.chunks.push(e)},he.prototype.onEnd=function(e){e===oe&&("string"===this.options.to?this.result=this.chunks.join(""):this.result=W(this.chunks)),this.chunks=[],this.err=e,this.msg=this.strm.msg};var we=he,be=ue,me=function(e,t){return(t=t||{}).raw=!0,ue(e,t)},ke=ue,_e=h,ge={Inflate:we,inflate:be,inflateRaw:me,ungzip:ke,constants:_e};e.Inflate=we,e.constants=_e,e.default=ge,e.inflate=be,e.inflateRaw=me,e.ungzip=ke,Object.defineProperty(e,"__esModule",{value:!0})}));

function fromZippedBase64(s) {
    let binary = atob(s);
    
    let bytesCompressed = new Uint8Array(binary.length);
    for (let i = 0; i < bytesCompressed.length; i++) {
        bytesCompressed[i] = binary.charCodeAt(i);
    }
    let inflated = pako.inflate(bytesCompressed);
    return inflated;
}


let size = 0;
let result = [];
for (const chunk of chunks) {
    const arr = fromZippedBase64(chunk);
    result.push(arr);
    size += arr.length;
}

let mergedArray = new Uint8Array(size);
let offset = 0;
for(const item of result){
    mergedArray.set(item, offset);
    offset += item.length;
}

return new TextDecoder().decode(mergedArray);
"""
                )

                stream.write(
                    index_html_contents[end_of_json_parse_index:].encode("utf-8")
                )



import json
import tempfile
from robocorp_ls_core.command_dispatcher import _CommandDispatcher
from robocorp_ls_core.python_ls import PythonLanguageServer
from robocorp_ls_core.basic import (
    overrides,
    log_and_silence_errors,
    log_but_dont_silence_errors,
)
import os
import time
from robotframework_ls.constants import (
    DEFAULT_COMPLETIONS_TIMEOUT,
    DEFAULT_COLLECT_DOCS_TIMEOUT,
    DEFAULT_LIST_TESTS_TIMEOUT,
)
from robocorp_ls_core.robotframework_log import get_logger
from typing import Any, Optional, Dict, Sequence, Set, ContextManager, Union
from robocorp_ls_core.protocols import (
    IConfig,
    IWorkspace,
    IIdMessageMatcher,
    IRobotFrameworkApiClient,
    IMonitor,
    IEndPoint,
    IProgressReporter,
    ActionResultDict,
    IFuture,
)
from pathlib import Path
from robotframework_ls.ep_providers import (
    EPConfigurationProvider,
    EPDirCacheProvider,
    EPEndPointProvider,
)
from robocorp_ls_core.jsonrpc.endpoint import require_monitor
from functools import partial
import itertools
from robotframework_ls import __version__, rf_interactive_integration
import typing
import sys
from robocorp_ls_core.watchdog_wrapper import IFSObserver
from robocorp_ls_core.lsp import (
    CodeLensTypedDict,
    TextDocumentPositionParamsTypedDict,
    PositionTypedDict,
    CompletionItemTypedDict,
    RenameParamsTypedDict,
    PrepareRenameParamsTypedDict,
    SelectionRangeParamsTypedDict,
    TextDocumentCodeActionTypedDict,
    WorkspaceEditParamsTypedDict,
    ShowDocumentParamsTypedDict,
)
from robotframework_ls.commands import (
    ROBOT_GET_RFLS_HOME_DIR,
    ROBOT_OPEN_FLOW_EXPLORER_INTERNAL,
    ROBOT_START_INDEXING_INTERNAL,
    ROBOT_WAIT_FULL_TEST_COLLECTION_INTERNAL,
    ROBOT_RF_INFO_INTERNAL,
    ROBOT_LINT_WORKSPACE,
    ROBOT_LINT_EXPLORER,
    ROBOT_GENERATE_FLOW_EXPLORER_MODEL,
    ROBOT_COLLECT_ROBOT_DOCUMENTATION,
    ROBOT_CONVERT_OUTPUT_XML_TO_ROBOSTREAM,
    ROBOT_APPLY_CODE_ACTION,
)
from robocorp_ls_core.jsonrpc.exceptions import JsonRpcException
import weakref
from io import StringIO


log = get_logger(__name__)

_LINT_DEBOUNCE_IN_SECONDS_LOW = 0.2
_LINT_DEBOUNCE_IN_SECONDS_HIGH = 0.8


class _CurrLintInfo(object):
    def __init__(
        self,
        rf_lint_api_client: IRobotFrameworkApiClient,
        lsp_messages,
        doc_uri,
        is_saved,
        on_finish,
    ) -> None:
        from robocorp_ls_core.lsp import LSPMessages
        from robocorp_ls_core.jsonrpc.monitor import Monitor

        self._rf_lint_api_client = rf_lint_api_client
        self.lsp_messages: LSPMessages = lsp_messages
        self.doc_uri = doc_uri
        self.is_saved = is_saved
        self._monitor = Monitor()
        self._on_finish = on_finish

    def __call__(self) -> None:
        from robocorp_ls_core.jsonrpc.exceptions import JsonRpcRequestCancelled
        from robocorp_ls_core.client_base import wait_for_message_matcher
        from robotframework_ls.server_api.client import SubprocessDiedError

        try:
            doc_uri = self.doc_uri
            self._monitor.check_cancelled()
            found = []
            message_matcher = self._rf_lint_api_client.request_lint(doc_uri)
            if message_matcher is not None:
                if wait_for_message_matcher(
                    message_matcher,
                    monitor=self._monitor,
                    request_cancel=self._rf_lint_api_client.request_cancel,
                    timeout=60 * 3,
                ):
                    diagnostics_msg = message_matcher.msg
                    if diagnostics_msg:
                        found = diagnostics_msg.get("result", [])
                    self._monitor.check_cancelled()
                    self.lsp_messages.publish_diagnostics(doc_uri, found)
        except JsonRpcRequestCancelled:
            log.debug("Cancelled linting: %s.", self.doc_uri)

        except SubprocessDiedError:
            log.debug("Subprocess exited while linting: %s.", self.doc_uri)

        except Exception:
            log.exception("Error linting: %s.", self.doc_uri)

        finally:
            self._on_finish(self)

    def cancel(self):
        self._monitor.cancel()


def run_in_new_thread(func, thread_name):
    import threading

    t = threading.Thread(target=func)
    t.name = thread_name
    t.start()


class _LintManager(object):
    def __init__(
        self, server_manager, lsp_messages, endpoint: IEndPoint, read_queue
    ) -> None:
        import threading
        import queue
        from robotframework_ls.server_manager import ServerManager

        self._server_manager: ServerManager = server_manager
        self._lsp_messages = lsp_messages
        self._endpoint = endpoint
        self._read_queue: queue.Queue = read_queue

        self._next_id = partial(next, itertools.count())

        self._lock = threading.Lock()
        self._doc_id_to_info: Dict[str, _CurrLintInfo] = {}  # requires lock
        self._uris_to_lint: Set[str] = set()  # requires lock

        self._progress_context: Optional[ContextManager[IProgressReporter]] = None
        self._progress_reporter: Optional[IProgressReporter] = None

    def schedule_lint(self, doc_uri: str, is_saved: bool, timeout: float) -> None:
        self.cancel_lint(doc_uri)

        # Note: this call must be done in the main thread.
        rf_lint_api_client = self._server_manager.get_lint_rf_api_client(doc_uri)
        if rf_lint_api_client is None:
            log.info("Unable to get lint api for: %s", doc_uri)
            return

        lock = self._lock
        doc_id_to_info = self._doc_id_to_info

        weak_self = weakref.ref(self)

        def on_finish(curr_info: _CurrLintInfo):
            from robocorp_ls_core.progress_report import progress_context

            # When it's finished, remove it from our references (if it's still
            # the current one...).
            #
            # Also, if we have remaining items which were manually scheduled,
            # we need to lint those.
            try:
                s = weak_self()
                next_uri_to_lint = None
                remaining_count = 0
                with lock:
                    if doc_id_to_info.get(curr_info.doc_uri) is curr_info:
                        del doc_id_to_info[curr_info.doc_uri]

                    if s is not None:
                        if self._progress_reporter is not None:
                            if self._progress_reporter.cancelled:
                                s._uris_to_lint.clear()

                        if s._uris_to_lint:
                            next_uri_to_lint = s._uris_to_lint.pop()
                            remaining_count = len(self._uris_to_lint)

                    if self._progress_context is None:
                        if remaining_count > 0:
                            self._progress_context = progress_context(
                                self._endpoint,
                                "Linting files... ",
                                None,
                                cancellable=True,
                            )
                            self._progress_reporter = self._progress_context.__enter__()
                    else:
                        if remaining_count == 0:
                            self._progress_context.__exit__(None, None, None)
                            self._progress_context = None
                            self._progress_reporter = None
                        else:
                            if self._progress_reporter is not None:
                                self._progress_reporter.set_additional_info(
                                    f"(remaining: {remaining_count})"
                                )

                if next_uri_to_lint and s is not None:
                    # As schedule lint must be done in the main thread, we
                    # we put an item in the queue to process the main thread.
                    def _schedule():
                        s = weak_self()
                        if s is not None:
                            s.schedule_lint(next_uri_to_lint, True, 0.0)

                    self._read_queue.put(_schedule)
            except:
                log.exception("Unhandled error on lint finish.")

        log.debug("Schedule lint for: %s", doc_uri)
        curr_info = _CurrLintInfo(
            rf_lint_api_client, self._lsp_messages, doc_uri, is_saved, on_finish
        )
        with self._lock:
            self._doc_id_to_info[doc_uri] = curr_info

        from robocorp_ls_core.timeouts import TimeoutTracker

        timeout_tracker = TimeoutTracker.get_singleton()
        timeout_tracker.call_on_timeout(
            timeout,
            partial(run_in_new_thread, curr_info, f"Lint: {doc_uri}"),
        )

    def cancel_lint(self, doc_uri: str) -> None:
        with self._lock:
            curr_info = self._doc_id_to_info.pop(doc_uri, None)
            if curr_info is not None:
                log.debug("Cancel lint for: %s", doc_uri)

                curr_info.cancel()

    def schedule_manual_lint(self, lint_paths: Sequence[str]) -> None:
        """
        This method is called to lint the given paths and provide the diagnostics
        as needed.

        It doesn't require files to be open and folders are recursively checked
        for files (.robot and .resource files).

        :param lint_paths: The paths that should be linted.
        """
        from robocorp_ls_core import uris

        new_uris_to_lint_set = set()

        for path in lint_paths:
            if path.lower().endswith((".robot", ".resource")):
                uri = uris.from_fs_path(str(path))
                new_uris_to_lint_set.add(uri)
                continue

            p = Path(path)
            if not p.is_dir():
                continue

            for f in p.rglob("*"):
                if f.suffix.lower() in (".robot", ".resource"):
                    uri = uris.from_fs_path(str(f))
                    new_uris_to_lint_set.add(uri)

        next_uri_to_lint = None
        with self._lock:
            self._uris_to_lint.update(new_uris_to_lint_set)
            if new_uris_to_lint_set:
                next_uri_to_lint = new_uris_to_lint_set.pop()

        if next_uri_to_lint:
            self.schedule_lint(next_uri_to_lint, True, 0.0)


command_dispatcher = _CommandDispatcher()


class RobotFrameworkLanguageServer(PythonLanguageServer):
    def __init__(self, rx, tx) -> None:
        from robocorp_ls_core.pluginmanager import PluginManager
        from robotframework_ls.rf_interactive_integration import _RfInterpretersManager
        from robotframework_ls.server_manager import ServerManager
        from robotframework_ls.ep_providers import DefaultConfigurationProvider
        from robotframework_ls.ep_providers import DefaultEndPointProvider
        from robotframework_ls.ep_providers import DefaultDirCacheProvider
        from robocorp_ls_core import watchdog_wrapper
        from robocorp_ls_core.remote_fs_observer_impl import RemoteFSObserver
        from robocorp_ls_core.options import Setup
        from robotframework_ls.robotframework_ls_completion_impl import (
            _RobotFrameworkLsCompletionImpl,
        )

        PythonLanguageServer.__init__(self, rx, tx)

        from robocorp_ls_core.cache import DirCache

        from robotframework_ls import robot_config

        home = robot_config.get_robotframework_ls_home()
        cache_dir = os.path.join(home, ".cache")

        log.debug(f"Cache dir: {cache_dir}")

        self._last_doc_uri: str = ""
        self._dir_cache = DirCache(cache_dir)

        self._pm = PluginManager()
        self._config_provider = DefaultConfigurationProvider(self.config)
        self._pm.set_instance(EPConfigurationProvider, self._config_provider)
        self._pm.set_instance(
            EPDirCacheProvider, DefaultDirCacheProvider(self._dir_cache)
        )
        self._pm.set_instance(
            EPEndPointProvider, DefaultEndPointProvider(self._endpoint)
        )

        weak_self = weakref.ref(self)

        def get_workspace_root_path():
            s = weak_self()  # We don't want a cyclic reference.
            if s is None:
                return None
            ws = s.workspace
            if ws is not None:
                return ws.root_path
            return None

        self._rf_interpreters_manager = _RfInterpretersManager(
            self._endpoint, self._pm, get_workspace_root_path=get_workspace_root_path
        )

        watch_impl = os.environ.get("ROBOTFRAMEWORK_LS_WATCH_IMPL", "auto")
        if watch_impl not in ("watchdog", "fsnotify", "auto"):
            log.info(
                f"ROBOTFRAMEWORK_LS_WATCH_IMPL should be 'auto', 'watchdog' or 'fsnotify'. Found: {watch_impl} (falling back to auto)"
            )
            watch_impl = "auto"

        if watch_impl == "auto":
            # In auto mode we use watchdog for windows and fsnotify (polling)
            # for Linux and Mac. The reason for that is that on Linux and Mac
            # if big folders are watched the system may complain due to the
            # lack of resources, which may prevent the extension from working
            # properly.
            #
            # If users want to opt-in, they can change to watchdog (and
            # ideally install it to their env to get native extensions).
            if sys.platform == "win32":
                watch_impl = "watchdog"
            else:
                watch_impl = "fsnotify"

        log.info(
            "Using watch implementation: %s (customize with ROBOTFRAMEWORK_LS_WATCH_IMPL environment variable)",
            watch_impl,
        )

        self._fs_observer = watchdog_wrapper.create_remote_observer(
            watch_impl, (".py", ".libspec", "robot", ".resource")
        )
        remote_observer = typing.cast(RemoteFSObserver, self._fs_observer)

        log_file = Setup.options.log_file
        verbose = Setup.options.verbose
        if not isinstance(log_file, str) and log_file is not None:
            log.critical("Expected %r to be a str (was: %s)", log_file, type(log_file))
            log_file = None

        if not isinstance(verbose, int) and verbose is not None:
            log.critical("Expected %r to be an int (was: %s)", verbose, type(verbose))
            verbose = None

        remote_observer.start_server(log_file=log_file, verbose=verbose)

        self._server_manager = ServerManager(self._pm, language_server=self)
        self._lint_manager = _LintManager(
            self._server_manager,
            self._lsp_messages,
            self._endpoint,
            self._jsonrpc_stream_reader.get_read_queue(),
        )
        self._robot_framework_ls_completion_impl = _RobotFrameworkLsCompletionImpl(
            self._server_manager, self
        )

    def _execute_on_main_thread(self, on_main_thread_callback):
        read_queue = self._jsonrpc_stream_reader.get_read_queue()
        read_queue.put(on_main_thread_callback)

    def get_remote_fs_observer_port(self) -> Optional[int]:
        from robocorp_ls_core.remote_fs_observer_impl import RemoteFSObserver

        remote_observer = typing.cast(RemoteFSObserver, self._fs_observer)
        return remote_observer.port

    @overrides(PythonLanguageServer._create_config)
    def _create_config(self) -> IConfig:
        from robotframework_ls.robot_config import RobotConfig

        return RobotConfig()

    @overrides(PythonLanguageServer._on_workspace_set)
    def _on_workspace_set(self, workspace: IWorkspace):
        PythonLanguageServer._on_workspace_set(self, workspace)
        self._server_manager.set_workspace(workspace)

    @overrides(PythonLanguageServer._obtain_fs_observer)
    def _obtain_fs_observer(self) -> IFSObserver:
        return self._fs_observer

    @overrides(PythonLanguageServer._create_workspace)
    def _create_workspace(
        self, root_uri: str, fs_observer: IFSObserver, workspace_folders
    ):
        from robotframework_ls.impl.robot_workspace import RobotWorkspace

        return RobotWorkspace(
            root_uri, fs_observer, workspace_folders, generate_ast=False
        )

    def m_initialize(
        self,
        processId=None,
        rootUri=None,
        rootPath=None,
        initializationOptions=None,
        workspaceFolders=None,
        **_kwargs,
    ) -> dict:
        # capabilities = _kwargs.get("capabilities", {})
        # text_document_capabilities = capabilities.get("textDocument", {})
        # document_symbol_capabilities = text_document_capabilities.get(
        #     "documentSymbol", {}
        # )
        # hierarchical_document_symbol_support = document_symbol_capabilities.get(
        #     "hierarchicalDocumentSymbolSupport", False
        # )
        # self._hierarchical_document_symbol_support = (
        #     hierarchical_document_symbol_support
        # )

        ret = PythonLanguageServer.m_initialize(
            self,
            processId=processId,
            rootUri=rootUri,
            rootPath=rootPath,
            initializationOptions=initializationOptions,
            workspaceFolders=workspaceFolders,
            **_kwargs,
        )

        initialization_options = initializationOptions
        if initialization_options:
            plugins_dir = initialization_options.get("pluginsDir")
            if isinstance(plugins_dir, str):
                if not os.path.isdir(plugins_dir):
                    log.critical(f"Expected: {plugins_dir} to be a directory.")
                else:
                    self._pm.load_plugins_from(Path(plugins_dir))

        return ret

    def forward_msg(self, msg: dict) -> None:
        method = msg["method"]
        self._endpoint.notify(method, msg["params"])

    @overrides(PythonLanguageServer.capabilities)
    def capabilities(self):
        from robocorp_ls_core.lsp import TextDocumentSyncKind
        from robotframework_ls.impl.semantic_tokens import TOKEN_TYPES, TOKEN_MODIFIERS
        from robotframework_ls import commands

        server_capabilities = {
            "codeActionProvider": {"resolveProvider": False},
            "codeLensProvider": {"resolveProvider": True},
            # Docs are lazily computed
            "completionProvider": {"resolveProvider": True},
            "documentFormattingProvider": True,
            "documentHighlightProvider": True,
            "documentRangeFormattingProvider": False,
            "documentSymbolProvider": True,
            "definitionProvider": True,
            "selectionRangeProvider": True,
            "executeCommandProvider": {
                "commands": [
                    "robot.addPluginsDir",
                    "robot.resolveInterpreter",
                    "robot.getLanguageServerVersion",
                    "robot.getInternalInfo",
                    "robot.listTests",
                ]
                + commands.ALL_SERVER_COMMANDS
            },
            "hoverProvider": True,
            "referencesProvider": True,
            "renameProvider": {
                "prepareProvider": True,
            },
            "foldingRangeProvider": True,
            # Note that there are no auto-trigger characters (there's no good
            # character as there's no `(` for parameters and putting it as a
            # space becomes a bit too much).
            # Under review: added ' ' and '\t' now that we have info on the
            # active parameter (waiting for feedback from community here to
            # check whether this is good or annoying).
            "signatureHelpProvider": {"triggerCharacters": [" ", "\t"]},
            "textDocumentSync": {
                "change": TextDocumentSyncKind.INCREMENTAL,
                "save": {"includeText": False},
                "openClose": True,
            },
            "workspace": {
                "workspaceFolders": {"supported": True, "changeNotifications": True}
            },
            "workspaceSymbolProvider": True,
            # The one below isn't accepted by lsp4j (it's still in LSP 3.15.0).
            # "workspaceSymbolProvider": {"workDoneProgress": False},
            "semanticTokensProvider": {
                "legend": {
                    "tokenTypes": TOKEN_TYPES,
                    "tokenModifiers": TOKEN_MODIFIERS,
                },
                "range": False,
                "full": True,
            },
        }
        log.debug("Server capabilities: %s", server_capabilities)
        return server_capabilities

    def m_workspace__execute_command(self, command: str = "", arguments=()) -> Any:
        if command.startswith("robot.internal.rfinteractive."):
            return rf_interactive_integration.execute_command(
                command, self, self._rf_interpreters_manager, arguments
            )

        return command_dispatcher.dispatch(self, command, arguments)

    @command_dispatcher("robot.addPluginsDir")
    def _add_plugins_dir(self, *arguments):
        directory: str = arguments[0]
        assert os.path.isdir(directory), f"Expected: {directory} to be a directory."
        self._pm.load_plugins_from(Path(directory))
        return True

    @command_dispatcher(ROBOT_COLLECT_ROBOT_DOCUMENTATION)
    def _collect_robot_documentation(
        self, opts
    ) -> Union[ActionResultDict, "partial[Any]"]:
        uri = opts.get("uri")
        if not uri:
            return dict(success=False, message="uri not provided", result=None)

        # We can either accept uri/library name or uri/line/col.
        # If both are provided the library_name is preferred.
        library_name = opts.get("library_name")
        line = opts.get("line")
        col = opts.get("col")
        if not library_name:
            if line is None or col is None:
                return dict(
                    success=False,
                    message="Either library_name or line/col must be provided.",
                    result=None,
                )

        rf_api_client = self._server_manager.get_others_api_client(uri)
        if rf_api_client is not None:
            func = partial(
                self._threaded_api_request_no_doc,
                rf_api_client,
                "request_collect_robot_documentation",
                __timeout__=DEFAULT_COLLECT_DOCS_TIMEOUT,
                doc_uri=uri,
                library_name=library_name,
                line=line,
                col=col,
            )
            func = require_monitor(func)
            return func

        return dict(success=False, message="no api available", result=None)

    @command_dispatcher(ROBOT_START_INDEXING_INTERNAL)
    def _start_indexing(self, *arguments):
        self._server_manager.get_regular_rf_api_client("")

    @command_dispatcher(ROBOT_LINT_WORKSPACE)
    def _lint_workspace(self, *arguments):
        folder_paths = self.workspace.get_folder_paths()
        self._lint_manager.schedule_manual_lint(folder_paths)

    @command_dispatcher(ROBOT_LINT_EXPLORER)
    def _lint_explorer(self, current_item, selection=None):
        if not selection:
            log.critical("The selection must be given for %s", ROBOT_LINT_EXPLORER)
            return

        if not isinstance(selection, (list, tuple)):
            log.critical(
                "The selection must be a list or tuple for %s. Found: %r",
                ROBOT_LINT_EXPLORER,
                selection,
            )
            return

        paths_to_analyze = set()
        for item in selection:
            if not isinstance(item, dict):
                log.critical(
                    "Expected item to be a dict for %s. Found: %r",
                    ROBOT_LINT_EXPLORER,
                    item,
                )
                continue

            fs_path = item.get("fsPath")
            if not fs_path:
                fs_path = item.get("path")
                if not fs_path:
                    log.critical(
                        "Expected item to be a dict with fsPath or path for %s. Found: %r",
                        ROBOT_LINT_EXPLORER,
                        item,
                    )
                    continue

            paths_to_analyze.add(fs_path)
        self._lint_manager.schedule_manual_lint(paths_to_analyze)

    def m_robot__provide_evaluatable_expression(
        self, uri: str, position: PositionTypedDict
    ):
        rf_api_client = self._server_manager.get_others_api_client(uri)
        if rf_api_client is not None:
            func = partial(
                self._threaded_api_request,
                rf_api_client,
                "request_evaluatable_expression",
                doc_uri=uri,
                position=position,
            )
            func = require_monitor(func)
            return func

        log.info("Unable to provide evaluatable expression (no api available).")
        return None

    @command_dispatcher(ROBOT_CONVERT_OUTPUT_XML_TO_ROBOSTREAM)
    def _convert_output_xml_to_rfstream(self, opts: Dict[str, Any]):
        def convert_in_thread():
            from robotframework_ls import import_robot_out_stream

            import_robot_out_stream()

            from robot_out_stream import xml_to_rfstream

            source = opts.get("xml_path")
            if source:
                with tempfile.NamedTemporaryFile(
                    "w", suffix=".robostream", delete=False
                ) as temp:

                    def write(contents):
                        temp.write(contents)

                    xml_to_rfstream.convert_xml_to_rfstream(source, write)

                # Note that we just send the temp file if an input file was received
                # (usually this should only happen if it's too big).
                # The client is responsible for deleting it after using it.
                return temp.name

            # We received contents (return contents too).
            xml_contents = opts.get("xml_contents")
            if xml_contents is None:
                raise RuntimeError(
                    f"Expected either xml_path or xml_contents. Keys: {list(opts.keys())}"
                )

            source = StringIO()
            source.write(xml_contents)
            source.seek(0)

            full_contents = []

            def write2(contents):
                full_contents.append(contents)

            xml_to_rfstream.convert_xml_to_rfstream(source, write2)
            return "".join(full_contents)

        return convert_in_thread

    @command_dispatcher(ROBOT_WAIT_FULL_TEST_COLLECTION_INTERNAL)
    def _wait_for_full_test_collection(self, *arguments):
        rf_api_client = self._server_manager.get_regular_rf_api_client("")
        if rf_api_client is not None:
            func = partial(
                self._threaded_api_request_no_doc,
                rf_api_client,
                "request_wait_for_full_test_collection",
            )
            func = require_monitor(func)
            return func

        log.info("Unable to wait for first test collection (no api available).")
        return []

    @command_dispatcher("robot.getInternalInfo")
    def _get_internal_info(self, *arguments):
        in_memory_docs = []
        workspace = self.workspace
        if workspace:
            for doc in workspace.iter_documents():
                in_memory_docs.append({"uri": doc.uri})
        return {
            "settings": self.config.get_full_settings(),
            "inMemoryDocs": in_memory_docs,
            "processId": os.getpid(),
        }

    @command_dispatcher("robot.resolveInterpreter")
    def _resolve_interpreter(self, *arguments):
        try:
            from robocorp_ls_core import uris
            from robotframework_ls.ep_resolve_interpreter import (
                EPResolveInterpreter,
            )
            from robotframework_ls.ep_resolve_interpreter import IInterpreterInfo

            target_robot = arguments[0]

            if isinstance(target_robot, (list, tuple)):
                if len(target_robot) > 0:
                    target_robot = target_robot[0]

            robot_uri = uris.from_fs_path(target_robot)

            for ep in self._pm.get_implementations(EPResolveInterpreter):
                interpreter_info: IInterpreterInfo = (
                    ep.get_interpreter_info_for_doc_uri(robot_uri)
                )
                if interpreter_info is not None:
                    return {
                        "pythonExe": interpreter_info.get_python_exe(),
                        "environ": interpreter_info.get_environ(),
                        "additionalPythonpathEntries": interpreter_info.get_additional_pythonpath_entries(),
                    }
        except:
            log.exception(f"Error resolving interpreter. Args: {arguments}")

    @command_dispatcher("robot.getLanguageServerVersion")
    def _get_language_server_version(self, *arguments):
        return __version__

    @command_dispatcher(ROBOT_GET_RFLS_HOME_DIR)
    def _get_rfls_home_dir(self, *arguments):
        from robotframework_ls import robot_config

        return robot_config.get_robotframework_ls_home()

    @command_dispatcher(ROBOT_RF_INFO_INTERNAL)
    def _get_rf_info_internal(self, *arguments):
        doc_uri = arguments[0]["uri"]

        api = self._server_manager._get_others_api(doc_uri)
        if api is not None:
            info = api.get_interpreter_info()
            rf_api_client = api.get_robotframework_api_client()
            if rf_api_client is not None:

                def func(monitor):
                    ret = self._threaded_api_request(
                        rf_api_client,
                        "request_rf_info",
                        doc_uri=doc_uri,
                        monitor=monitor,
                    )
                    if info is not None:
                        if isinstance(ret, dict):
                            ret["interpreter_id"] = info.get_interpreter_id()
                    return ret

                func = require_monitor(func)
                return func

        log.info("Unable to get RF info (no api available).")
        return None

    @command_dispatcher(ROBOT_GENERATE_FLOW_EXPLORER_MODEL)
    def _generate_flow_explorer_model(self, opts: Dict[str, Any]):
        """
        :param opts:
            A dictionary with the options to generate the flow explorer model.
            Options available:
                - uri: target uri mapping to the robot or directory with multiple
                  robots from where the model should be generated.

        """
        uri = opts["uri"]

        return self.async_api_forward("request_flow_explorer_model", "api", doc_uri=uri)

    @command_dispatcher(ROBOT_OPEN_FLOW_EXPLORER_INTERNAL)
    def _open_flow_explorer(self, opts) -> ActionResultDict:
        """
        :param dict opts:
            Keys:
                "currentFileUri": uri for target (may also be folder).
                "htmlBundleFolderPath":folder which contains the html/js assets.
        """
        from robotframework_ls.constants import (
            DEFAULT_ROBOT_FLOW_EXPLORER_HTML_TEMPLATE,
            DEFAULT_ROBOT_FLOW_EXPLORER_OPTIONS,
        )
        from robocorp_ls_core.uris import from_fs_path
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_FLOW_EXPLORER_THEME_DARK,
        )
        from robotframework_ls.impl.robot_generated_lsp_constants import (
            OPTION_ROBOT_FLOW_EXPLORER_THEME,
        )

        current_doc_uri = opts["currentFileUri"]
        html_bundle_flow_folder_path = opts["htmlBundleFolderPath"]
        log.info(
            "Dispatched flow explorer internal command with args:",
            str(current_doc_uri),
            str(html_bundle_flow_folder_path),
        )

        def finish_and_convert_result(result: ActionResultDict) -> ActionResultDict:
            log.info("Generating web page for visualization...")

            if not result["success"]:
                return result

            # identify the selected theme & construct options
            rfe_theme = self._config.get_setting(
                OPTION_ROBOT_FLOW_EXPLORER_THEME,
                str,
                OPTION_ROBOT_FLOW_EXPLORER_THEME_DARK,
            )
            rfe_options = DEFAULT_ROBOT_FLOW_EXPLORER_OPTIONS.copy()
            rfe_options["theme"] = rfe_theme
            model = result["result"]

            # filling in the HTML template
            # adding the new model as data to Flow Explorer
            replacement_html = DEFAULT_ROBOT_FLOW_EXPLORER_HTML_TEMPLATE.substitute(
                rfe_options=json.dumps(rfe_options),
                rfe_data=json.dumps(model),
                rfe_favicon_path=Path(
                    os.path.join(html_bundle_flow_folder_path, "favicon.png")
                ).as_uri(),
                rfe_js_path=Path(
                    os.path.join(
                        html_bundle_flow_folder_path, "robot_flow_explorer_bundle.js"
                    )
                ).as_uri(),
            )
            # create the temporary HTML file to display
            with tempfile.NamedTemporaryFile(
                mode="w", delete=False, suffix=".html"
            ) as temp_file:
                Path(temp_file.name).write_text(replacement_html, encoding="utf-8")
                return {
                    "result": from_fs_path(temp_file.name),
                    "success": True,
                    "message": None,
                }

        return self.async_api_forward(
            "request_flow_explorer_model",
            "api",
            doc_uri=current_doc_uri,
            uri=current_doc_uri,
            __add_doc_uri_in_args__=False,
            __convert_result__=finish_and_convert_result,
        )

    @command_dispatcher("robot.listTests")
    def _list_tests(self, *arguments):
        doc_uri = arguments[0]["uri"]

        rf_api_client = self._server_manager.get_others_api_client(doc_uri)
        if rf_api_client is not None:
            func = partial(
                self._threaded_api_request_no_doc,
                rf_api_client,
                "request_list_tests",
                doc_uri=doc_uri,
                __timeout__=DEFAULT_LIST_TESTS_TIMEOUT,
            )
            func = require_monitor(func)
            return func

        log.info("Unable to list tests (no api available).")
        return []

    @overrides(PythonLanguageServer.m_workspace__did_change_configuration)
    @log_and_silence_errors(log)
    def m_workspace__did_change_configuration(self, **kwargs) -> None:
        from robotframework_ls.impl import robot_localization

        PythonLanguageServer.m_workspace__did_change_configuration(self, **kwargs)
        robot_localization.set_global_from_config(self.config)
        self._server_manager.set_config(self.config)

    # --- Methods to forward to the api

    @overrides(PythonLanguageServer.m_shutdown)
    @log_and_silence_errors(log)
    def m_shutdown(self, **kwargs):
        try:
            from robocorp_ls_core.remote_fs_observer_impl import RemoteFSObserver

            remote_observer = typing.cast(RemoteFSObserver, self._fs_observer)
            remote_observer.dispose()
        except Exception:
            log.exception("Error disposing RemoteFSObserver.")
        self._server_manager.shutdown()

        PythonLanguageServer.m_shutdown(self, **kwargs)

    @overrides(PythonLanguageServer.m_exit)
    @log_and_silence_errors(log)
    def m_exit(self, **kwargs):
        self._server_manager.exit()

        PythonLanguageServer.m_exit(self, **kwargs)

    def m_text_document__formatting(
        self, textDocument=None, options=None
    ) -> Optional[list]:
        doc_uri = textDocument["uri"]

        source_format_rf_api_client = self._server_manager.get_others_api_client(
            doc_uri
        )
        if source_format_rf_api_client is None:
            log.info("Unable to get API for source format.")
            return []

        message_matcher = source_format_rf_api_client.request_source_format(
            text_document=textDocument, options=options
        )
        if message_matcher is None:
            raise RuntimeError(
                "Error requesting code formatting (message_matcher==None)."
            )
        curtime = time.time()
        maxtime = curtime + DEFAULT_COMPLETIONS_TIMEOUT

        # i.e.: wait X seconds for the code format and bail out if we
        # can't get it.
        available_time = maxtime - time.time()
        if available_time <= 0:
            raise RuntimeError("Code formatting timed-out (available_time <= 0).")

        if message_matcher.event.wait(available_time):
            msg = message_matcher.msg
            if msg is not None:
                result = msg.get("result")
                if result:
                    return result
                else:
                    return []
        raise RuntimeError("Code formatting timed-out.")

    @overrides(PythonLanguageServer.m_text_document__did_close)
    def m_text_document__did_close(self, textDocument=None, **_kwargs):
        self._server_manager.forward(
            ("api", "lint", "others"),
            "textDocument/didClose",
            {"textDocument": textDocument},
        )
        PythonLanguageServer.m_text_document__did_close(
            self, textDocument=textDocument, **_kwargs
        )

    @overrides(PythonLanguageServer.m_text_document__did_open)
    def m_text_document__did_open(self, textDocument=None, **_kwargs):
        self._server_manager.forward(
            ("api", "lint", "others"),
            "textDocument/didOpen",
            {"textDocument": textDocument},
        )
        PythonLanguageServer.m_text_document__did_open(
            self, textDocument=textDocument, **_kwargs
        )

    @overrides(PythonLanguageServer.m_text_document__did_change)
    def m_text_document__did_change(
        self, contentChanges=None, textDocument=None, **_kwargs
    ):
        self._server_manager.forward(
            ("api", "lint", "others"),
            "textDocument/didChange",
            {"contentChanges": contentChanges, "textDocument": textDocument},
        )
        PythonLanguageServer.m_text_document__did_change(
            self, contentChanges=contentChanges, textDocument=textDocument, **_kwargs
        )

    @overrides(PythonLanguageServer.m_workspace__did_change_workspace_folders)
    def m_workspace__did_change_workspace_folders(self, event=None, **_kwargs):
        self._server_manager.forward(
            ("api", "lint", "others"),
            "workspace/didChangeWorkspaceFolders",
            {"event": event},
        )
        PythonLanguageServer.m_workspace__did_change_workspace_folders(
            self, event=event, **_kwargs
        )

    # --- Customized implementation

    @overrides(PythonLanguageServer.lint)
    def lint(self, doc_uri, is_saved, content_changes=None) -> None:
        # We want the timeout to be shorter if it was saved or if the user
        # typed a new line.
        timeout = _LINT_DEBOUNCE_IN_SECONDS_LOW
        if not is_saved:
            timeout = _LINT_DEBOUNCE_IN_SECONDS_HIGH
            for change in content_changes:
                try:
                    text = change.get("text", "")
                    if "\n" in text or "\r" in text:
                        timeout = _LINT_DEBOUNCE_IN_SECONDS_LOW
                        break
                except:
                    log.exception(
                        "Error computing lint timeout with: %s", content_changes
                    )
        self._lint_manager.schedule_lint(doc_uri, is_saved, timeout)

    @overrides(PythonLanguageServer.cancel_lint)
    def cancel_lint(self, doc_uri) -> None:
        self._lint_manager.cancel_lint(doc_uri)

    def m_completion_item__resolve(self, **params):
        completion_item: CompletionItemTypedDict = params
        return self._robot_framework_ls_completion_impl.resolve_completion_item(
            completion_item
        )

    def m_text_document__completion(self, **params):
        doc_uri = params["textDocument"]["uri"]
        # Note: 0-based
        line, col = params["position"]["line"], params["position"]["character"]
        return self._robot_framework_ls_completion_impl.text_document_completion(
            doc_uri, line, col
        )

    @log_but_dont_silence_errors(log)
    def _threaded_api_request(
        self,
        rf_api_client: IRobotFrameworkApiClient,
        request_method_name: str,
        doc_uri: str,
        monitor: IMonitor,
        __timeout__=DEFAULT_COMPLETIONS_TIMEOUT,
        __log__=False,
        __convert_result__=None,
        **kwargs,
    ):
        from robocorp_ls_core.client_base import wait_for_message_matcher

        func = getattr(rf_api_client, request_method_name)

        ws = self.workspace
        if not ws:
            log.critical(
                "Workspace must be set before calling %s.", request_method_name
            )
            return None

        document = ws.get_document(doc_uri, accept_from_file=True)
        if document is None:
            log.critical(
                "Unable to find document (%s) for %s." % (doc_uri, request_method_name)
            )
            return None

        # Asynchronous completion.
        message_matcher: Optional[IIdMessageMatcher] = func(doc_uri, **kwargs)
        if message_matcher is None:
            log.debug("Message matcher for %s returned None.", request_method_name)
            return None

        if wait_for_message_matcher(
            message_matcher,
            rf_api_client.request_cancel,
            __timeout__,
            monitor,
        ):
            msg = message_matcher.msg
            if msg is not None:
                result = msg.get("result")
                if __log__:
                    log.info("Result: %s", result)
                if result:
                    if __convert_result__ is not None:
                        result = __convert_result__(result)
                    return result

                error = msg.get("error")
                if error:
                    raise JsonRpcException(**error)

        return None

    @log_but_dont_silence_errors(log)
    def _threaded_api_request_no_doc(
        self,
        rf_api_client: IRobotFrameworkApiClient,
        request_method_name: str,
        monitor: Optional[IMonitor],
        __timeout__=DEFAULT_COMPLETIONS_TIMEOUT,
        __convert_result__=None,
        **kwargs,
    ):
        from robocorp_ls_core.client_base import wait_for_message_matcher

        func = getattr(rf_api_client, request_method_name)

        # Asynchronous completion.
        message_matcher: Optional[IIdMessageMatcher] = func(**kwargs)
        if message_matcher is None:
            log.debug("Message matcher for %s returned None.", request_method_name)
            return None

        if wait_for_message_matcher(
            message_matcher,
            rf_api_client.request_cancel,
            __timeout__,
            monitor,
        ):
            msg = message_matcher.msg
            if msg is not None:
                result = msg.get("result")
                if result:
                    if __convert_result__ is not None:
                        result = __convert_result__(result)
                    return result

        return None

    def async_api_forward(
        self,
        api_client_method_name: str,
        target_api: str,  # api, lint, others
        doc_uri: str,
        default_return=None,
        __add_doc_uri_in_args__=True,
        **kwargs,
    ):
        rf_api_client: Optional[IRobotFrameworkApiClient]
        if target_api == "api":
            rf_api_client = self._server_manager.get_regular_rf_api_client(doc_uri)
        elif target_api == "lint":
            rf_api_client = self._server_manager.get_lint_rf_api_client(doc_uri)
        elif target_api == "others":
            rf_api_client = self._server_manager.get_others_api_client(doc_uri)
        else:
            raise AssertionError(f"Unexpected: {target_api}")

        self._last_doc_uri = doc_uri
        if rf_api_client is not None:
            if __add_doc_uri_in_args__:
                func = partial(
                    self._threaded_api_request,
                    rf_api_client,
                    api_client_method_name,
                    doc_uri=doc_uri,
                    **kwargs,
                )
            else:
                func = partial(
                    self._threaded_api_request_no_doc,
                    rf_api_client,
                    api_client_method_name,
                    **kwargs,
                )
            func = require_monitor(func)
            return func

        log.info(
            "No api available (call: %s, uri: %s).", api_client_method_name, doc_uri
        )
        return default_return

    def m_text_document__definition(self, **kwargs):
        doc_uri = kwargs["textDocument"]["uri"]
        # Note: 0-based
        line, col = kwargs["position"]["line"], kwargs["position"]["character"]
        return self.async_api_forward(
            "request_find_definition", "api", doc_uri, line=line, col=col
        )

    def m_text_document__rename(self, **kwargs):
        rename_params: RenameParamsTypedDict = kwargs
        # Note: 0-based
        doc_uri: str = rename_params["textDocument"]["uri"]
        line: int = rename_params["position"]["line"]
        col: int = rename_params["position"]["character"]
        new_name = rename_params["newName"]
        return self.async_api_forward(
            "request_rename", "api", doc_uri, line=line, col=col, new_name=new_name
        )

    def m_text_document__prepare_rename(self, **kwargs):
        rename_params: PrepareRenameParamsTypedDict = kwargs
        # Note: 0-based
        doc_uri: str = rename_params["textDocument"]["uri"]
        line: int = rename_params["position"]["line"]
        col: int = rename_params["position"]["character"]
        return self.async_api_forward(
            "request_prepare_rename", "api", doc_uri, line=line, col=col
        )

    def m_text_document__signature_help(self, **kwargs):
        """
        "params": {
            "textDocument": {
                "uri": "file:///x%3A/vscode-robot/local_test/Basic/resources/keywords.robot"
            },
            "position": {"line": 7, "character": 22},
            "context": {
                "isRetrigger": False,
                "triggerCharacter": " ",
                "triggerKind": 2,
            },
        },
        """
        doc_uri = kwargs["textDocument"]["uri"]
        # Note: 0-based
        line, col = kwargs["position"]["line"], kwargs["position"]["character"]

        return self.async_api_forward(
            "request_signature_help", "api", doc_uri, line=line, col=col
        )

    def m_text_document__folding_range(self, **kwargs):
        """
        "params": {
            "textDocument": {
                "uri": "file:///x%3A/vscode-robot/local_test/Basic/resources/keywords.robot"
            },
        },
        """
        doc_uri = kwargs["textDocument"]["uri"]

        return self.async_api_forward("request_folding_range", "others", doc_uri)

    def m_text_document__selection_range(self, **kwargs):
        params: SelectionRangeParamsTypedDict = kwargs
        # Note: 0-based
        doc_uri: str = params["textDocument"]["uri"]
        positions = params["positions"]
        return self.async_api_forward(
            "request_selection_range", "others", doc_uri, positions=positions
        )

    def m_text_document__code_lens(self, **kwargs):
        doc_uri = kwargs["textDocument"]["uri"]
        return self.async_api_forward("request_code_lens", "others", doc_uri)

    def m_code_lens__resolve(self, **kwargs):
        code_lens: CodeLensTypedDict = kwargs

        code_lens_command = code_lens.get("command")
        data = code_lens.get("data")
        if code_lens_command is None and isinstance(data, dict):
            # For the interactive shell we need to resolve the arguments.
            doc_uri = data.get("uri")

            return self.async_api_forward(
                "request_resolve_code_lens",
                "others",
                doc_uri,
                code_lens=code_lens,
                __add_doc_uri_in_args__=False,
            )

        return code_lens

    def m_text_document__document_symbol(self, **kwargs):
        doc_uri = kwargs["textDocument"]["uri"]
        return self.async_api_forward("request_document_symbol", "others", doc_uri)

    def m_text_document__hover(self, **kwargs):
        params: TextDocumentPositionParamsTypedDict = kwargs
        doc_uri = params["textDocument"]["uri"]
        # Note: 0-based
        line, col = params["position"]["line"], params["position"]["character"]

        return self.async_api_forward(
            "request_hover", "api", doc_uri, line=line, col=col
        )

    def m_text_document__references(self, **kwargs):
        doc_uri = kwargs["textDocument"]["uri"]
        # Note: 0-based
        line, col = kwargs["position"]["line"], kwargs["position"]["character"]
        include_declaration = kwargs["context"]["includeDeclaration"]

        # Note: we want to use the same one used by m_workspace__symbol (to reuse
        # the related caches).
        rf_api_client = self._server_manager.get_regular_rf_api_client(doc_uri)
        if rf_api_client is not None:
            func = partial(
                self._threaded_api_request,
                rf_api_client,
                "request_references",
                doc_uri=doc_uri,
                line=line,
                col=col,
                include_declaration=include_declaration,
                __timeout__=9999999,
            )
            func = require_monitor(func)
            return func

        log.info("Unable to compute references (no api available).")
        return []

    def m_text_document__semantic_tokens__range(self, textDocument=None, range=None):
        raise RuntimeError("Not currently implemented!")

    def m_text_document__semantic_tokens__full(self, textDocument=None):
        doc_uri = textDocument["uri"]

        return self.async_api_forward(
            "request_semantic_tokens_full",
            "others",
            doc_uri,
            default_return={"resultId": None, "data": []},
            text_document=textDocument,
            __add_doc_uri_in_args__=False,
        )

    def m_workspace__symbol(self, query: Optional[str] = None) -> Any:
        doc_uri = self._last_doc_uri
        return self.async_api_forward(
            "request_workspace_symbols",
            "api",
            doc_uri,
            query=query,
            __add_doc_uri_in_args__=False,
        )

    def m_text_document__document_highlight(self, **kwargs):
        params: TextDocumentPositionParamsTypedDict = kwargs
        doc_uri = params["textDocument"]["uri"]
        line, col = params["position"]["line"], params["position"]["character"]

        return self.async_api_forward(
            "request_document_highlight", "others", doc_uri, line=line, col=col
        )

    def m_cancel_progress(self, progressId):
        if not PythonLanguageServer.m_cancel_progress(self, progressId=progressId):
            for api in self._server_manager.collect_apis():
                # We don't keep track of which server started which progress,
                # so, we send it to all the APIs.
                api.forward_async("cancelProgress", {"progressId": progressId})
        return True

    def m_text_document__code_action(self, **kwargs):
        params: TextDocumentCodeActionTypedDict = kwargs
        # Sample params:
        # {
        #     "textDocument": {
        #         "uri": "file:///x%3A/vscode-robot/local_test/robot_check/checkmy.robot"
        #     },
        #     "range": {
        #         "start": {"line": 12, "character": 5},
        #         "end": {"line": 12, "character": 5},
        #     },
        #     "context": {
        #         "diagnostics": [
        #             {
        #                 "range": {
        #                     "start": {"line": 12, "character": 4},
        #                     "end": {"line": 12, "character": 8},
        #                 },
        #                 "message": "Undefined keyword: rara.",
        #                 "severity": 1,
        #                 "source": "robotframework",
        #             }
        #         ],
        #         "triggerKind": 1,
        #     },
        # }
        doc_uri = params["textDocument"]["uri"]
        # Note: we have to use the "api" server which has symbols caches.
        return self.async_api_forward(
            "request_code_action", "api", doc_uri, params=params
        )

    @command_dispatcher(ROBOT_APPLY_CODE_ACTION)
    def _apply_code_action(self, code_action_info):
        weak_self = weakref.ref(self)

        def in_thread(*args, **kwargs):
            self = weak_self()
            if self is None:
                return

            apply_edit: WorkspaceEditParamsTypedDict = code_action_info["apply_edit"]
            fut: IFuture = self._lsp_messages.apply_edit_args(apply_edit)
            try:
                fut.result(4)
            except:
                log.exception(f"Exception calling: {self._lsp_messages.M_APPLY_EDIT}")

            # Deal with linting (because if only a dependent file is changed we
            # still want to ask for the main file to be linted).
            lint_uris = code_action_info.get("lint_uris")
            if lint_uris:
                for uri in lint_uris:
                    self._execute_on_main_thread(partial(self.lint, uri, is_saved=True))

            # Force showing some document?
            show_document_args: Optional[
                ShowDocumentParamsTypedDict
            ] = code_action_info.get("show_document")
            log.info("Show document: %s", show_document_args)
            self._lsp_messages.show_document(show_document_args)

        return in_thread
